# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Coordinate systems that describe how data & model states are discretized."""

from __future__ import annotations

import dataclasses
import math
from typing import Any, Iterable

from dinosaur import coordinate_systems as dinosaur_coordinates
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import vertical_interpolation
import jax
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import typing
import numpy as np
from penzai.core import struct


SphericalHarmonicsImpl = spherical_harmonic.SphericalHarmonicsImpl
RealSphericalHarmonics = spherical_harmonic.RealSphericalHarmonics
FastSphericalHarmonics = spherical_harmonic.FastSphericalHarmonics
P = jax.sharding.PartitionSpec


@dataclasses.dataclass
class ArrayKey:
  """Wrapper for a numpy array to make it hashable."""

  value: np.ndarray

  def __eq__(self, other):
    return (
        isinstance(self, ArrayKey)
        and self.value.dtype == other.value.dtype
        and self.value.shape == other.value.shape
        and (self.value == other.value).all()
    )

  def __ne__(self, other):
    return not self == other

  def __hash__(self) -> int:
    return hash((self.value.shape, self.value.tobytes()))


@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass
class TimeDelta(cx.Coordinate):
  """Coordinates that discretize data along static relative time."""

  time: np.ndarray = dataclasses.field(metadata={'pytree_node': False})
  offset: float = dataclasses.field(
      default=0.0, metadata={'pytree_node': False}
  )

  @property
  def dims(self):
    return ('timedelta',)

  @property
  def shape(self):
    return self.time.shape

  @property
  def fields(self):
    return {'timedelta': cx.wrap(self.time, self)}

  @classmethod
  def as_index(cls, axis_size: int, offset: float = 0.0) -> TimeDelta:
    return cls(time=np.arange(axis_size), offset=offset)

  def _components(self):
    return (self.offset, ArrayKey(self.time))

  def tree_flatten(self):
    """Flattens TimeDelta."""
    aux_data = self._components()
    return (), aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, leaves):
    """Unflattens TimeDelta."""
    del leaves  # unused
    offset, array_key = aux_data
    return cls(offset=offset, time=array_key.value)

  def __eq__(self, other):
    return (
        isinstance(other, TimeDelta)
        and self._components() == other._components()
    )

  def __ne__(self, other):
    return not self == other

  def __hash__(self) -> int:
    return hash(self._components())


#
# Grid-like and spherical harmonic coordinate systems
#


# TODO(dkochkov) Consider leaving out spherical_harmonics_impl from repr.
@struct.pytree_dataclass
class LonLatGrid(cx.Coordinate, struct.Struct):
  """Coordinates that discretize data as point values on lon-lat grid."""

  longitude_nodes: int = dataclasses.field(metadata={'pytree_node': False})
  latitude_nodes: int = dataclasses.field(metadata={'pytree_node': False})
  latitude_spacing: str = dataclasses.field(
      default='gauss', metadata={'pytree_node': False}
  )
  longitude_offset: float = dataclasses.field(
      default=0.0, metadata={'pytree_node': False}
  )
  radius: float | None = dataclasses.field(
      default=None, metadata={'pytree_node': False}
  )
  spherical_harmonics_impl: SphericalHarmonicsImpl = dataclasses.field(
      default=FastSphericalHarmonics,
      kw_only=True,
      metadata={'pytree_node': False},
  )
  longitude_wavenumbers: int = dataclasses.field(
      default=0, repr=False, kw_only=True, metadata={'pytree_node': False}
  )
  total_wavenumbers: int = dataclasses.field(
      default=0, repr=False, kw_only=True, metadata={'pytree_node': False}
  )
  _ylm_grid: spherical_harmonic.Grid = dataclasses.field(
      init=False, repr=False, metadata={'pytree_node': False}
  )

  def __post_init__(self):
    ylm_grid = spherical_harmonic.Grid(
        longitude_wavenumbers=self.longitude_wavenumbers,
        total_wavenumbers=self.total_wavenumbers,
        longitude_nodes=self.longitude_nodes,
        latitude_nodes=self.latitude_nodes,
        latitude_spacing=self.latitude_spacing,
        longitude_offset=self.longitude_offset,
        radius=self.radius,
        spherical_harmonics_impl=self.spherical_harmonics_impl,
    )
    object.__setattr__(self, '_ylm_grid', ylm_grid)

  @property
  def ylm_grid(self) -> spherical_harmonic.Grid:
    return self._ylm_grid

  @property
  def dims(self):
    return ('longitude', 'latitude')

  @property
  def shape(self):
    return self.ylm_grid.nodal_shape

  @property
  def fields(self):
    return {
        k: cx.wrap(v, cx.SelectedAxis(self, i))
        for i, (k, v) in enumerate(zip(self.dims, self.ylm_grid.nodal_axes))
    }

  def to_spherical_harmonic_grid(
      self,
      longitude_wavenumbers: int | None = None,
      total_wavenumbers: int | None = None,
      spherical_harmonics_impl: SphericalHarmonicsImpl | None = None,
  ):
    """Constructs a `SphericalHarmonicGrid` from the `LonLatGrid`."""
    if longitude_wavenumbers is None:
      longitude_wavenumbers = self.longitude_wavenumbers
    if total_wavenumbers is None:
      total_wavenumbers = self.total_wavenumbers
    if spherical_harmonics_impl is None:
      spherical_harmonics_impl = self.spherical_harmonics_impl
    return SphericalHarmonicGrid(
        longitude_wavenumbers=longitude_wavenumbers,
        total_wavenumbers=total_wavenumbers,
        longitude_offset=self.longitude_offset,
        radius=self.radius,
        spherical_harmonics_impl=spherical_harmonics_impl,
        longitude_nodes=self.longitude_nodes,
        latitude_nodes=self.latitude_nodes,
        latitude_spacing=self.latitude_spacing,
    )

  @classmethod
  def from_dinosaur_grid(
      cls,
      ylm_grid: spherical_harmonic.Grid,
  ):
    return cls(
        longitude_nodes=ylm_grid.longitude_nodes,
        latitude_nodes=ylm_grid.latitude_nodes,
        latitude_spacing=ylm_grid.latitude_spacing,
        longitude_offset=ylm_grid.longitude_offset,
        radius=ylm_grid.radius,
        total_wavenumbers=ylm_grid.total_wavenumbers,
        longitude_wavenumbers=ylm_grid.longitude_wavenumbers,
        spherical_harmonics_impl=ylm_grid.spherical_harmonics_impl,
    )

  @classmethod
  def construct(
      cls,
      max_wavenumber: int,
      gaussian_nodes: int,
      latitude_spacing: str = 'gauss',
      longitude_offset: float = 0.0,
      radius: float | None = None,
      spherical_harmonics_impl: SphericalHarmonicsImpl = FastSphericalHarmonics,
  ) -> LonLatGrid:
    """Constructs a `LonLatGrid` compatible with max_wavenumber and nodes.

    Args:
      max_wavenumber: maximum wavenumber to resolve.
      gaussian_nodes: number of nodes on the Gaussian grid between the equator
        and a pole.
      latitude_spacing: either 'gauss' or 'equiangular'. This determines the
        spacing of nodal grid points in the latitudinal (north-south) direction.
      longitude_offset: the value of the first longitude node, in radians.
      radius: radius of the sphere. If `None` a default values of `1` is used.
      spherical_harmonics_impl: class providing an implementation of spherical
        harmonics.

    Returns:
      Constructed LonLatGrid object.
    """
    ylm_grid = spherical_harmonic.Grid(
        longitude_wavenumbers=max_wavenumber + 1,
        total_wavenumbers=max_wavenumber + 2,
        longitude_nodes=4 * gaussian_nodes,
        latitude_nodes=2 * gaussian_nodes,
        latitude_spacing=latitude_spacing,
        longitude_offset=longitude_offset,
        spherical_harmonics_impl=spherical_harmonics_impl,
        radius=radius,
    )
    return cls.from_dinosaur_grid(ylm_grid=ylm_grid)

  # The factory methods below return "standard" grids that appear in the
  # literature. See, e.g. https://doi.org/10.5194/tc-12-1499-2018 and
  # https://www.ecmwf.int/en/forecasts/documentation-and-support/data-spatial-coordinate-systems

  # The number in these names correspond to the maximum resolved wavenumber,
  # which is one less than the number of wavenumbers used in the Grid
  # constructor. An additional total wavenumber is added because the top
  # wavenumber is clipped from the initial state and each calculation of
  # explicit tendencies.

  # The names for these factory methods (including capilatization) are
  # standard in the literature.
  # pylint:disable=invalid-name

  # T* grids can model quadratic terms without aliasing, because the maximum
  # total wavenumber is <= 2/3 of the number of latitudinal nodes. ECMWF
  # sometimes calls these "TQ" (truncated quadratic) grids.

  @classmethod
  def T21(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=21, gaussian_nodes=16, **kwargs)

  @classmethod
  def T31(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=31, gaussian_nodes=24, **kwargs)

  @classmethod
  def T42(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=42, gaussian_nodes=32, **kwargs)

  @classmethod
  def T85(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=85, gaussian_nodes=64, **kwargs)

  @classmethod
  def T106(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=106, gaussian_nodes=80, **kwargs)

  @classmethod
  def T119(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=119, gaussian_nodes=90, **kwargs)

  @classmethod
  def T170(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=170, gaussian_nodes=128, **kwargs)

  @classmethod
  def T213(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=213, gaussian_nodes=160, **kwargs)

  @classmethod
  def T340(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=340, gaussian_nodes=256, **kwargs)

  @classmethod
  def T425(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=425, gaussian_nodes=320, **kwargs)

  # TL* grids do not truncate any frequencies, and hence can only model linear
  # terms exactly. ECMWF used "TL" (truncated linear) grids for semi-Lagrangian
  # advection (which eliminates quadratic terms) up to 2016, which it switched
  # to "cubic" grids for resolutions above TL1279:
  # https://www.ecmwf.int/sites/default/files/elibrary/2016/17262-new-grid-ifs.pdf

  @classmethod
  def TL31(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=31, gaussian_nodes=16, **kwargs)

  @classmethod
  def TL47(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=47, gaussian_nodes=24, **kwargs)

  @classmethod
  def TL63(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=63, gaussian_nodes=32, **kwargs)

  @classmethod
  def TL95(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=95, gaussian_nodes=48, **kwargs)

  @classmethod
  def TL127(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=127, gaussian_nodes=64, **kwargs)

  @classmethod
  def TL159(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=159, gaussian_nodes=80, **kwargs)

  @classmethod
  def TL179(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=179, gaussian_nodes=90, **kwargs)

  @classmethod
  def TL255(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=255, gaussian_nodes=128, **kwargs)

  @classmethod
  def TL639(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=639, gaussian_nodes=320, **kwargs)

  @classmethod
  def TL1279(cls, **kwargs) -> LonLatGrid:
    return cls.construct(max_wavenumber=1279, gaussian_nodes=640, **kwargs)


@struct.pytree_dataclass
class SphericalHarmonicGrid(cx.Coordinate, struct.Struct):
  """Coordinates that discretize data as spherical harmonic coefficients."""

  longitude_wavenumbers: int = dataclasses.field(
      metadata={'pytree_node': False}
  )
  total_wavenumbers: int = dataclasses.field(metadata={'pytree_node': False})
  longitude_offset: float = dataclasses.field(
      default=0.0, metadata={'pytree_node': False}
  )
  radius: float | None = dataclasses.field(
      default=None, metadata={'pytree_node': False}
  )
  spherical_harmonics_impl: SphericalHarmonicsImpl = dataclasses.field(
      default=FastSphericalHarmonics,
      kw_only=True,
      metadata={'pytree_node': False},
  )
  longitude_nodes: int = dataclasses.field(
      default=0, repr=False, kw_only=True, metadata={'pytree_node': False}
  )
  latitude_nodes: int = dataclasses.field(
      default=0, repr=False, kw_only=True, metadata={'pytree_node': False}
  )
  latitude_spacing: str = dataclasses.field(
      default='gauss', repr=False, kw_only=True, metadata={'pytree_node': False}
  )
  _ylm_grid: spherical_harmonic.Grid = dataclasses.field(
      init=False, repr=False, metadata={'pytree_node': False}
  )

  def __post_init__(self):
    ylm_grid = spherical_harmonic.Grid(
        longitude_wavenumbers=self.longitude_wavenumbers,
        total_wavenumbers=self.total_wavenumbers,
        longitude_offset=self.longitude_offset,
        radius=self.radius,
        spherical_harmonics_impl=self.spherical_harmonics_impl,
        longitude_nodes=self.longitude_nodes,
        latitude_nodes=self.latitude_nodes,
        latitude_spacing=self.latitude_spacing,
    )
    object.__setattr__(self, '_ylm_grid', ylm_grid)

  @property
  def ylm_grid(self) -> spherical_harmonic.Grid:
    return self._ylm_grid

  @property
  def dims(self):
    return ('longitude_wavenumber', 'total_wavenumber')

  @property
  def shape(self) -> tuple[int, ...]:
    return self.ylm_grid.modal_shape

  @property
  def fields(self):
    return {
        k: cx.wrap(v, cx.SelectedAxis(self, i))
        for i, (k, v) in enumerate(zip(self.dims, self.ylm_grid.modal_axes))
    }

  def to_lon_lat_grid(
      self,
      longitude_nodes: int | None = None,
      latitude_nodes: int | None = None,
      latitude_spacing: str | None = None,
  ):
    """Constructs a `LonLatGrid` from the `SphericalHarmonicGrid`."""
    if longitude_nodes is None:
      longitude_nodes = self.longitude_nodes
    if latitude_nodes is None:
      latitude_nodes = self.latitude_nodes
    if latitude_spacing is None:
      latitude_spacing = self.latitude_spacing
    return LonLatGrid(
        longitude_nodes=longitude_nodes,
        latitude_nodes=latitude_nodes,
        latitude_spacing=latitude_spacing,
        longitude_offset=self.longitude_offset,
        radius=self.radius,
        spherical_harmonics_impl=self.spherical_harmonics_impl,
        longitude_wavenumbers=self.longitude_wavenumbers,
        total_wavenumbers=self.total_wavenumbers,
    )

  @classmethod
  def from_dinosaur_grid(
      cls,
      ylm_grid: spherical_harmonic.Grid,
  ):
    return cls(
        longitude_wavenumbers=ylm_grid.longitude_wavenumbers,
        total_wavenumbers=ylm_grid.total_wavenumbers,
        longitude_offset=ylm_grid.longitude_offset,
        radius=ylm_grid.radius,
        spherical_harmonics_impl=ylm_grid.spherical_harmonics_impl,
        longitude_nodes=ylm_grid.longitude_nodes,
        latitude_nodes=ylm_grid.latitude_nodes,
        latitude_spacing=ylm_grid.latitude_spacing,
    )

  @classmethod
  def with_wavenumbers(
      cls,
      longitude_wavenumbers: int,
      dealiasing: str = 'quadratic',
      latitude_spacing: str = 'gauss',
      longitude_offset: float = 0.0,
      spherical_harmonics_impl: SphericalHarmonicsImpl = (
          FastSphericalHarmonics
      ),
      radius: float | None = None,
  ) -> SphericalHarmonicGrid:
    """Constructs a `SphericalHarmonicGrid` by specifying only wavenumbers."""
    # The number of nodes is chosen for de-aliasing.
    order = {'linear': 2, 'quadratic': 3, 'cubic': 4}[dealiasing]
    longitude_nodes = order * longitude_wavenumbers + 1
    latitude_nodes = math.ceil(longitude_nodes / 2)
    ylm_grid = spherical_harmonic.Grid(
        longitude_wavenumbers=longitude_wavenumbers,
        total_wavenumbers=longitude_wavenumbers + 1,
        longitude_nodes=longitude_nodes,
        latitude_nodes=latitude_nodes,
        latitude_spacing=latitude_spacing,
        longitude_offset=longitude_offset,
        spherical_harmonics_impl=spherical_harmonics_impl,
        radius=radius,
    )
    return cls.from_dinosaur_grid(ylm_grid=ylm_grid)

  @classmethod
  def construct(
      cls,
      max_wavenumber: int,
      gaussian_nodes: int,
      latitude_spacing: str = 'gauss',
      longitude_offset: float = 0.0,
      radius: float | None = None,
      spherical_harmonics_impl: SphericalHarmonicsImpl = (
          FastSphericalHarmonics
      ),
  ) -> SphericalHarmonicGrid:
    """Constructs a `SphericalHarmonicGrid` with max_wavenumber.

    Args:
      max_wavenumber: maximum wavenumber to resolve.
      gaussian_nodes: number of nodes on the Gaussian grid between the equator
        and a pole.
      latitude_spacing: either 'gauss' or 'equiangular'. This determines the
        spacing of nodal grid points in the latitudinal (north-south) direction.
      longitude_offset: the value of the first longitude node, in radians.
      radius: radius of the sphere. If `None` a default values of `1` is used.
      spherical_harmonics_impl: class providing an implementation of spherical
        harmonics.

    Returns:
      Constructed SphericalHarmonicGrid object.
    """
    ylm_grid = spherical_harmonic.Grid(
        longitude_wavenumbers=max_wavenumber + 1,
        total_wavenumbers=max_wavenumber + 2,
        longitude_nodes=4 * gaussian_nodes,
        latitude_nodes=2 * gaussian_nodes,
        latitude_spacing=latitude_spacing,
        longitude_offset=longitude_offset,
        spherical_harmonics_impl=spherical_harmonics_impl,
        radius=radius,
    )
    return cls.from_dinosaur_grid(ylm_grid=ylm_grid)

  # The factory methods below return "standard" grids that appear in the
  # literature. See, e.g. https://doi.org/10.5194/tc-12-1499-2018 and
  # https://www.ecmwf.int/en/forecasts/documentation-and-support/data-spatial-coordinate-systems

  # The number in these names correspond to the maximum resolved wavenumber,
  # which is one less than the number of wavenumbers used in the Grid
  # constructor. An additional total wavenumber is added because the top
  # wavenumber is clipped from the initial state and each calculation of
  # explicit tendencies.

  # The names for these factory methods (including capilatization) are
  # standard in the literature.
  # pylint:disable=invalid-name

  # T* grids can model quadratic terms without aliasing, because the maximum
  # total wavenumber is <= 2/3 of the number of latitudinal nodes. ECMWF
  # sometimes calls these "TQ" (truncated quadratic) grids.

  @classmethod
  def T21(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=21, gaussian_nodes=16, **kwargs)

  @classmethod
  def T31(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=31, gaussian_nodes=24, **kwargs)

  @classmethod
  def T42(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=42, gaussian_nodes=32, **kwargs)

  @classmethod
  def T85(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=85, gaussian_nodes=64, **kwargs)

  @classmethod
  def T106(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=106, gaussian_nodes=80, **kwargs)

  @classmethod
  def T119(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=119, gaussian_nodes=90, **kwargs)

  @classmethod
  def T170(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=170, gaussian_nodes=128, **kwargs)

  @classmethod
  def T213(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=213, gaussian_nodes=160, **kwargs)

  @classmethod
  def T340(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=340, gaussian_nodes=256, **kwargs)

  @classmethod
  def T425(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=425, gaussian_nodes=320, **kwargs)

  # TL* grids do not truncate any frequencies, and hence can only model linear
  # terms exactly. ECMWF used "TL" (truncated linear) grids for semi-Lagrangian
  # advection (which eliminates quadratic terms) up to 2016, which it switched
  # to "cubic" grids for resolutions above TL1279:
  # https://www.ecmwf.int/sites/default/files/elibrary/2016/17262-new-grid-ifs.pdf

  @classmethod
  def TL31(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=31, gaussian_nodes=16, **kwargs)

  @classmethod
  def TL47(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=47, gaussian_nodes=24, **kwargs)

  @classmethod
  def TL63(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=63, gaussian_nodes=32, **kwargs)

  @classmethod
  def TL95(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=95, gaussian_nodes=48, **kwargs)

  @classmethod
  def TL127(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=127, gaussian_nodes=64, **kwargs)

  @classmethod
  def TL159(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=159, gaussian_nodes=80, **kwargs)

  @classmethod
  def TL179(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=179, gaussian_nodes=90, **kwargs)

  @classmethod
  def TL255(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=255, gaussian_nodes=128, **kwargs)

  @classmethod
  def TL639(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=639, gaussian_nodes=320, **kwargs)

  @classmethod
  def TL1279(cls, **kwargs) -> SphericalHarmonicGrid:
    return cls.construct(max_wavenumber=1279, gaussian_nodes=640, **kwargs)


#
# Vertical level coordinates
#


@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass
class SigmaLevels(cx.Coordinate):
  """Coordinates that discretize data as fraction of the surface pressure."""

  boundaries: np.ndarray = dataclasses.field(metadata={'pytree_node': False})
  sigma_levels: sigma_coordinates.SigmaCoordinates = dataclasses.field(
      init=False,
      repr=False,
      compare=False,
      metadata={'pytree_node': False},
  )

  def __init__(self, boundaries: Iterable[float] | np.ndarray):
    boundaries = np.asarray(boundaries)
    self.boundaries = boundaries
    self.__post_init__()

  def __post_init__(self):
    sigma_levels = sigma_coordinates.SigmaCoordinates(
        boundaries=self.boundaries
    )
    object.__setattr__(self, 'sigma_levels', sigma_levels)

  @property
  def dims(self):
    return ('sigma',)

  @property
  def shape(self) -> tuple[int, ...]:
    return self.sigma_levels.centers.shape

  @property
  def fields(self):
    return {'sigma': cx.wrap(self.sigma_levels.centers, self)}

  def asdict(self) -> dict[str, Any]:
    return {k: v.tolist() for k, v in dataclasses.asdict(self).items()}

  def _components(self):
    return (ArrayKey(self.boundaries),)

  def tree_flatten(self):
    """Flattens SigmaLevels."""
    aux_data = self._components()
    return (), aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, leaves):
    """Unflattens SigmaLevels."""
    del leaves  # unused
    (boundaries_key,) = aux_data
    return cls(boundaries=boundaries_key.value)

  def __eq__(self, other):
    return (
        isinstance(other, SigmaLevels)
        and self._components() == other._components()
    )

  def __ne__(self, other):
    return not self == other

  def __hash__(self) -> int:
    return hash(self._components())

  @classmethod
  def from_dinosaur_sigma_levels(
      cls,
      sigma_levels: sigma_coordinates.SigmaCoordinates,
  ):
    return cls(boundaries=sigma_levels.boundaries)

  @classmethod
  def equidistant(
      cls,
      layers: int,
  ) -> SigmaLevels:
    sigma_levels = sigma_coordinates.SigmaCoordinates.equidistant(layers)
    boundaries = sigma_levels.boundaries
    return cls(boundaries=boundaries)


@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass
class PressureLevels(cx.Coordinate):
  """Coordinates that discretize data per pressure levels."""

  centers: np.ndarray = dataclasses.field(metadata={'pytree_node': False})
  pressure_levels: vertical_interpolation.PressureCoordinates = (
      dataclasses.field(
          init=False,
          repr=False,
          compare=False,
          metadata={'pytree_node': False},
      )
  )

  def __init__(self, centers: Iterable[float] | np.ndarray):
    centers = np.asarray(centers)
    self.centers = centers
    self.__post_init__()

  def __post_init__(self):
    pressure_levels = vertical_interpolation.PressureCoordinates(
        centers=self.centers
    )
    object.__setattr__(self, 'pressure_levels', pressure_levels)

  @property
  def dims(self):
    return ('pressure',)

  @property
  def shape(self) -> tuple[int, ...]:
    return self.centers.shape

  @property
  def fields(self):
    return {'pressure': cx.wrap(self.centers, self)}

  def asdict(self) -> dict[str, Any]:
    return {k: v.tolist() for k, v in dataclasses.asdict(self).items()}

  def _components(self):
    return (ArrayKey(self.centers),)

  def tree_flatten(self):
    """Flattens PressureLevels."""
    aux_data = self._components()
    return (), aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, leaves):
    """Unflattens PressureLevels."""
    del leaves  # unused
    (centers,) = aux_data
    return cls(centers=centers.value)

  def __eq__(self, other):
    return (
        isinstance(other, PressureLevels)
        and self._components() == other._components()
    )

  def __ne__(self, other):
    return not self == other

  def __hash__(self) -> int:
    return hash(self._components())

  @classmethod
  def from_dinosaur_pressure_levels(
      cls,
      pressure_levels: vertical_interpolation.PressureCoordinates,
  ):
    return cls(centers=pressure_levels.centers)


@struct.pytree_dataclass
class LayerLevels(cx.Coordinate, struct.Struct):
  """Coordinates that discretize data by index of unstructured layer."""

  n_layers: int = dataclasses.field(metadata={'pytree_node': False})

  @property
  def dims(self):
    return ('layer_index',)

  @property
  def shape(self) -> tuple[int, ...]:
    return (self.n_layers,)

  @property
  def fields(self):
    return {'layer_index': cx.wrap(np.arange(self.n_layers), self)}


#
# Solver-specific coordinate combinations
#


@struct.pytree_dataclass
class DinosaurCoordinates(cx.CartesianProduct, struct.Struct):
  """Coordinate that is product of horizontal & vertical coorinates.

  This combined coordinate object is useful for compactly keeping track of the
  full coordinate system of the Dinosaur dynamic core or pressure-level
  representation of the spherical shell data.
  """

  coordinates: tuple[cx.Coordinate, ...] = dataclasses.field(
      init=False, metadata={'pytree_node': False}
  )
  horizontal: LonLatGrid | SphericalHarmonicGrid = dataclasses.field(
      metadata={'pytree_node': False}
  )
  vertical: SigmaLevels | PressureLevels | LayerLevels = dataclasses.field(
      metadata={'pytree_node': False}
  )
  dycore_partition_spec: jax.sharding.PartitionSpec = dataclasses.field(
      metadata={'pytree_node': False}, default=P('z', 'x', 'y')
  )
  physics_partition_spec: jax.sharding.PartitionSpec = dataclasses.field(
      metadata={'pytree_node': False}, default=P(None, ('x', 'z'), 'y')
  )

  def __init__(
      self,
      horizontal,
      vertical,
      dycore_partition_spec: jax.sharding.PartitionSpec = P('z', 'x', 'y'),
      physics_partition_spec: jax.sharding.PartitionSpec = P(
          None, ('x', 'z'), 'y'
      ),
  ):
    super().__init__(coordinates=(vertical, horizontal))
    self.horizontal = horizontal
    self.vertical = vertical
    self.dycore_partition_spec = dycore_partition_spec
    self.physics_partition_spec = physics_partition_spec

  @property
  def dims(self):
    return self.vertical.dims + self.horizontal.dims

  @property
  def shape(self) -> tuple[int, ...]:
    return self.vertical.shape + self.horizontal.shape

  @property
  def fields(self):
    return self.vertical.fields | self.horizontal.fields

  @property
  def dinosaur_coords(self):
    """Returns the CoordinateSystem object from the Dinosaur package."""
    # TODO(dkochkov) Either make spmd_mesh an argument or ideally add
    # ShardingMesh object to the new API to hold sharding information.
    spmd_mesh = None  # make this an argument and change property to a method.
    horizontal, vertical = self.horizontal, self.vertical
    horizontal = horizontal.ylm_grid
    if isinstance(vertical, SigmaLevels):
      vertical = vertical.sigma_levels
    elif isinstance(vertical, PressureLevels):
      vertical = vertical.pressure_levels
    elif isinstance(vertical, LayerLevels):
      pass
    else:
      raise ValueError(f'Unsupported vertical {vertical=}')
    return dinosaur_coordinates.CoordinateSystem(
        horizontal=horizontal, vertical=vertical, spmd_mesh=spmd_mesh
    )

  @property
  def dinosaur_grid(self):
    return self.dinosaur_coords.horizontal

  @classmethod
  def from_dinosaur_coords(
      cls,
      coords: dinosaur_coordinates.CoordinateSystem,
  ):
    """Constructs instance from coordinates in Dinosaur package."""
    horizontal = LonLatGrid.from_dinosaur_grid(coords.horizontal)
    if isinstance(coords.vertical, sigma_coordinates.SigmaCoordinates):
      vertical = SigmaLevels.from_dinosaur_sigma_levels(coords.vertical)
    elif isinstance(
        coords.vertical, vertical_interpolation.PressureCoordinates
    ):
      vertical = PressureLevels.from_dinosaur_pressure_levels(coords.vertical)
    else:
      raise ValueError(f'Unsupported vertical {coords.vertical=}')
    return cls(horizontal=horizontal, vertical=vertical)


#
# Helper functions.
#
# TODO(dkochkov) Refactor/remove helpers below to coordax.coords.


def consistent_coords(*inputs) -> cx.Coordinate:
  """Returns the unique coordinate, or raises a ValueError."""
  if not all([cx.is_positional_prefix_field(f) for f in inputs]):
    raise ValueError(f'{inputs=} are not in positional prefix order.')
  dim_names = {tuple(f.named_shape.keys()) for f in inputs}
  if len(dim_names) != 1:
    raise ValueError(f'Found non-unique {dim_names=} in inputs.')
  (dim_names,) = dim_names
  coords = {
      cx.compose_coordinates(*[f.coords[k] for k in dim_names]) for f in inputs
  }
  if len(coords) != 1:
    raise ValueError(f'Found non-unique {coords=} in inputs.')
  (coords,) = coords
  return coords


def modal_shape_to_nodal_shape(
    modal_shape: typing.Pytree,
    grid: SphericalHarmonicGrid,
) -> typing.Pytree:
  """Returns the nodal shape corresponding to the given modal shape."""
  # Note: here modal_shape is expected to be a pytree with
  # typing.ShapeDtypeStruct leaves, so that eval_shape works as expected.
  to_nodal_shape = lambda x: jax.eval_shape(grid.ylm_grid.to_nodal, x)
  return jax.tree.map(to_nodal_shape, modal_shape)


def nodal_shape_to_modal_shape(
    nodal_shape: typing.Pytree,
    grid: SphericalHarmonicGrid,
) -> typing.Pytree:
  """Returns the modal shape corresponding to the given nodal shape."""
  # Note: here nodal_shape is expected to be a pytree with
  # typing.ShapeDtypeStruct leaves, so that eval_shape works as expected.
  to_modal_shape = lambda x: jax.eval_shape(grid.ylm_grid.to_modal, x)
  return jax.tree.map(to_modal_shape, nodal_shape)


def split_field_attrs(pytree):
  """Splits pytree of `Field` into data, sim_time and spec."""
  is_field = lambda x: isinstance(x, cx.Field)
  fields, treedef = jax.tree.flatten(pytree, is_leaf=is_field)
  data = jax.tree.unflatten(treedef, [x.data for x in fields])
  coords = consistent_coords(*fields)
  return data, coords



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines `filtering` that aim to improve stability of integration."""

from typing import Any, Callable, Dict, Optional, Sequence, Union
from dinosaur import coordinate_systems
from dinosaur import filtering
from dinosaur import pytree_utils
from dinosaur import scales
from dinosaur import time_integration
from dinosaur import typing
import gin
import haiku as hk
import jax
import numpy as np


QuantityOrStr = Union[str, scales.Quantity]

StepFilterModule = Callable[..., typing.PyTreeStepFilterFn]
TransformModule = typing.TransformModule


#  =============================================================================
#  Step filters that attenuate modal components between time steps.
#  =============================================================================


@gin.register
class NoFilter(hk.Module):
  """Filter module that performs no filtering."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    del u  # unused.
    return u_next


@gin.register
class ClipFilter(hk.Module):
  """Filter that clips highest total wavenumber in the next state."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      wavenumbers_to_clip: int = 1,
      name: Optional[str] = None,
  ):
    """See `time_integration.exponential_filter` for details."""
    del dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.coords = coords
    self.wavenumbers_to_clip = wavenumbers_to_clip

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    del u  # unused.
    return self.coords.horizontal.clip_wavenumbers(
        u_next, self.wavenumbers_to_clip
    )


@gin.register
class ExponentialLeapfrogFilter(hk.Module):
  """Filter that removes high frequency components from a spectral state."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      tau: QuantityOrStr = '0.010938',
      order: int = 18,
      cutoff: float = 0,
      name: Optional[str] = None,
  ):
    """See `time_integration.exponential_filter` for details."""
    del aux_features  # unused.
    super().__init__(name=name)
    tau = physics_specs.nondimensionalize(scales.Quantity(tau))
    self.filter_fn = time_integration.exponential_leapfrog_step_filter(
        coords.horizontal, dt, tau, order, cutoff)

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    return self.filter_fn(u, u_next)


@gin.register
class ExponentialFilter(hk.Module):
  """Filter that removes high frequency components from a spectral state."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      tau: QuantityOrStr = '0.010938',
      order: int = 18,
      cutoff: float = 0,
      name: Optional[str] = None,
  ):
    """See `time_integration.exponential_step_filter` for details."""
    del aux_features  # unused.
    super().__init__(name=name)
    tau = physics_specs.nondimensionalize(scales.Quantity(tau))
    self.filter_fn = time_integration.exponential_step_filter(
        coords.horizontal, dt, tau, order, cutoff)

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    return self.filter_fn(u, u_next)


@gin.register
class HorizontalDiffusionFilter(hk.Module):
  """Filter that applies implicit diffusion operator to a spectral state."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      tau: QuantityOrStr = '',
      order: int = 1,
      name: Optional[str] = None,
  ):
    """See `time_integration.horizontal_diffusion_filter` for details."""
    del aux_features  # unused.
    super().__init__(name=name)
    tau = physics_specs.nondimensionalize(scales.Quantity(tau))
    self.filter_fn = time_integration.horizontal_diffusion_step_filter(
        coords.horizontal, dt, tau, order)

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    del u  # unused
    return self.filter_fn(u_next)  # pytype: disable=wrong-arg-count  # always-use-return-annotations


@gin.register
class RobertAsselinLeapfrogFilter(hk.Module):
  """Time smoothing filter."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      strength: float = 0.05,
      name: Optional[str] = None,
  ):
    """See `time_integration.robert_asselin_leapfrog_filter` for details."""
    del dt, coords, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.filter_fn = time_integration.robert_asselin_leapfrog_filter(strength)

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    return self.filter_fn(u, u_next)


@gin.register
class LearnedExponentialFilter(hk.Module):
  """Low pass filter with learned parameters."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      name: Optional[str] = None,
  ):
    del dt, physics_specs, aux_features  # unused.
    self.coords = coords
    self.a_init = hk.initializers.Constant(16)
    self.p_init = hk.initializers.Constant(18)
    self.c_init = hk.initializers.Constant(0)
    super().__init__(name=name)

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    del u  # unused.
    a_logit = hk.get_parameter('attenuation_logit', shape=(), init=self.a_init)
    p_logit = hk.get_parameter('order_logit', shape=(), init=self.p_init)
    c_logit = hk.get_parameter('threshold_logit', shape=(), init=self.c_init)
    a = jax.nn.softplus(a_logit)
    p = jax.nn.softplus(p_logit)
    c = jax.nn.sigmoid(c_logit)
    filter_fn = filtering.exponential_filter(self.coords.horizontal, a, p, c)
    return filter_fn(u_next)


@gin.register
class SequentialStepFilter(hk.Module):
  """Filter module that combines multiple step filters applied sequentially."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      filter_modules: Sequence[StepFilterModule],
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.filter_fns = [module(coords, dt, physics_specs, aux_features)
                       for module in filter_modules]

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    for filter_fn in self.filter_fns:
      u_next = filter_fn(u, u_next)
    return u_next


@gin.register
class LayeredStepFilter(hk.Module):
  """Filter decorator that uses varying time-scales at different levels."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      filter_module: Union[HorizontalDiffusionFilter, ExponentialFilter],
      tau_vals: Union[Sequence[float], np.ndarray],
      tau_units: QuantityOrStr,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    tau = (np.asarray(tau_vals) * scales.Quantity(tau_units))
    tau = tau[:, np.newaxis, np.newaxis]  # add spatial axes.
    self.filter_fn = filter_module(
        coords, dt, physics_specs, aux_features, tau=tau)

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    return self.filter_fn(u, u_next)


@gin.register
class MaskedFilter(hk.Module):
  """Filter that is only applied to a part of the state."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      filter_module: StepFilterModule,
      mask: typing.Pytree,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.filter_fn = filter_module(coords, dt, physics_specs, aux_features)
    self.mask = mask

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    mask = type(u_next)(**self.mask)  # convert to same structure.
    return jax.tree_util.tree_map(
        lambda x, y, b: self.filter_fn(x, y) if b else y,
        u, u_next, mask)


@gin.register
class FilterFromTransform(hk.Module):
  """Filter module that wraps a transform module."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      transform_module: TransformModule,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.transform_fn = transform_module(
        coords, dt, physics_specs, aux_features)

  def __call__(
      self,
      u: typing.PyTreeState,
      u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    return self.transform_fn(u_next)


@gin.register
class FixGlobalMeanFilter(hk.Module):
  """Filter that removes the change in the global mean of certain keys."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      keys: tuple[str, ...] = ('log_surface_pressure',),
      name: Optional[str] = None,
  ):
    del aux_features  # unused.
    super().__init__(name=name)
    self.keys = keys

  def __call__(
      self, u: typing.PyTreeState, u_next: typing.PyTreeState
  ) -> typing.PyTreeState:
    u_dict, _ = pytree_utils.as_dict(u)
    u_dict, _ = pytree_utils.flatten_dict(u_dict)
    u_next_dict, from_dict_fn = pytree_utils.as_dict(u_next)
    u_next_dict, _ = pytree_utils.flatten_dict(u_next_dict)
    for key in self.keys:
      global_mean = u_dict[key][..., 0]
      u_next_dict[key] = u_next_dict[key].at[..., 0].set(global_mean)
    u_next_dict = pytree_utils.unflatten_dict(u_next_dict)
    return from_dict_fn(u_next_dict)


#  =============================================================================
#  Filters that act on modal variables without time-step context.
#  =============================================================================


@gin.register
class DataNoFilter(hk.Module):
  """Filter module that performs no filtering."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    return inputs


@gin.register
class DataExponentialFilter(hk.Module):
  """Filter that removes high frequency components from a modal data."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      attenuation: float = 16,
      order: int = 18,
      cutoff: float = 0,
      name: Optional[str] = None,
  ):
    """See `filtering.exponential_filter` for details."""
    del dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.filter_fn = filtering.exponential_filter(
        coords.horizontal, attenuation, order, cutoff)

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    return self.filter_fn(inputs)


@gin.register
class PerVariableDataFilter(hk.Module):
  """Filter module that applies different filters for each variable."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      per_variable_filters: Dict[str, Any],
      name: Optional[str] = None,
  ):
    """See `filtering.exponential_filter` for details."""
    super().__init__(name=name)
    self.filter_fns = jax.tree_util.tree_map(
        lambda m: m(coords, dt, physics_specs, aux_features),
        per_variable_filters)

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    inputs_dict, from_dict_fn = pytree_utils.as_dict(inputs)
    return from_dict_fn(jax.tree_util.tree_map(
        lambda x, fn: fn(x), inputs_dict, self.filter_fns))



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for parameterization modules in gcm.ml.parameterizations.py."""

from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import coordinate_systems
from dinosaur import primitive_equations_states
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
import gin
import haiku as hk
import jax

from neuralgcm import features  # pylint: disable=unused-import
from neuralgcm import parameterizations  # pylint: disable=unused-import
from neuralgcm import physics_specifications
from neuralgcm import transforms  # pylint: disable=unused-import

SCALE = scales.DEFAULT_SCALE
units = scales.units


@gin.register
class ParameterizationTowerMock(hk.Module):
  """Tower that returns random outputs of expected shape."""

  def __init__(self, output_size, *args):
    super().__init__(name=None)
    self.output_size = output_size

  def __call__(self, inputs):
    del inputs  # unused.
    return jax.random.uniform(
        jax.random.PRNGKey(42),
        shape=[self.output_size, 1, 1])  # include 2d spatial dimensions


class ParameterizationsTest(parameterized.TestCase):
  """Tests parameterization modules."""

  @parameterized.parameters(
      dict(
          module_name='DirectNeuralParameterization',
          features_name='PrimitiveEquationsDiagnosticState',),
      dict(
          module_name='DivCurlNeuralParameterization',
          features_name='PrimitiveEquationsDiagnosticState',),
  )
  def test_parameterization_output_shapes(self, module_name, features_name):
    """Tests that parameterizations produce outputs with expected shapes."""
    gin.enter_interactive_mode()
    gin.clear_config()
    physics_gin_config = '\n'.join([
        'get_physics_specs.construct_fn = @primitive_eq_specs_constructor',
    ])
    gin.parse_config(physics_gin_config)
    physics_specs = physics_specifications.get_physics_specs()
    dt = physics_specs.nondimensionalize(100 * units.s)
    # setting up coordinates and initial state.
    grid = spherical_harmonic.Grid.with_wavenumbers(32)
    vertical_grid = sigma_coordinates.SigmaCoordinates.equidistant(4)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    initial_state_fn, aux_features = (
        primitive_equations_states.isothermal_rest_atmosphere(
            coords, physics_specs))

    @gin.configurable
    def parameterization_fwd(x, parameterization_cls):
      return parameterization_cls(coords, dt, physics_specs, aux_features)(x)

    parameterization_config = '\n'.join([
        f'parameterization_fwd.parameterization_cls = @{module_name}',
        f'{module_name}.modal_to_nodal_features_module = @{features_name}',
        f'{features_name}.features_transform_module = @IdentityTransform',
        f'{module_name}.nodal_mapping_module = @NodalMapping',
        f'{module_name}.tendency_transform_module = @IdentityTransform',
        'NodalMapping.tower_factory = @ParameterizationTowerMock'
    ])
    gin.parse_config(parameterization_config)
    inputs = initial_state_fn(jax.random.PRNGKey(42))
    model = hk.without_apply_rng(hk.transform(parameterization_fwd))
    params = model.init(jax.random.PRNGKey(42), inputs)
    predicted_tendencies = model.apply(params, inputs)
    for actual, expected in zip(jax.tree_util.tree_leaves(predicted_tendencies),
                                jax.tree_util.tree_leaves(inputs)):
      self.assertEqual(actual.shape, expected.shape)


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines ``Field`` and base ``Coordinate`` classes that define coordax API.

``Coordinate`` objects define a discretization schema, dimension names and
provide methods & coordinate field values to facilitate computations.
``Field`` objects keep track of positional and named dimensions of an array.
Named dimensions of a ``Field`` are associated with coordinates that describe
their discretization.

Current implementation of ``Field`` and associated helper methods like `cmap`
(coordinate-map) are based on `penzai.core.named_axes` library.
"""

from __future__ import annotations

import abc
import collections
import dataclasses
import functools
import operator
from typing import Any, Callable, Hashable, Mapping, Sequence, TypeAlias, TypeGuard

import jax
import jax.numpy as jnp
import numpy as np
from penzai.core import named_axes
from penzai.core import struct


AxisName: TypeAlias = Hashable
Pytree: TypeAlias = Any


class Coordinate(abc.ABC):
  """Abstract class for coordinate objects.

  Coordinate subclasses are expected to obey several invariants:
  1. Dimension names may not be repeated: `len(set(dims)) == len(dims)`
  2. All dimensions must be named: `len(shape) == len(dims)`
  """

  @property
  @abc.abstractmethod
  def dims(self) -> tuple[AxisName, ...]:
    """Dimension names of the coordinate."""
    raise NotImplementedError()

  @property
  @abc.abstractmethod
  def shape(self) -> tuple[int, ...]:
    """Shape of the coordinate."""
    raise NotImplementedError()

  @property
  @abc.abstractmethod
  def fields(self) -> dict[AxisName, Field]:
    """A maps from field names to their values."""

  @property
  def sizes(self) -> dict[AxisName, int]:
    """Sizes of all dimensions on this coordinate."""
    return dict(zip(self.dims, self.shape))

  @property
  def ndim(self) -> int:
    """Dimensionality of the coordinate."""
    return len(self.dims)


@dataclasses.dataclass
class ArrayKey:
  """Wrapper for a numpy array to make it hashable."""

  value: np.ndarray

  def __eq__(self, other):
    return (
        isinstance(self, ArrayKey)
        and self.value.dtype == other.value.dtype
        and self.value.shape == other.value.shape
        and (self.value == other.value).all()
    )

  def __ne__(self, other):
    return not self == other

  def __hash__(self) -> int:
    return hash((self.value.shape, self.value.tobytes()))


@struct.pytree_dataclass
class SelectedAxis(Coordinate, struct.Struct):
  """Coordinate that exposes one dimension of a multidimensional coordinate."""

  coordinate: Coordinate
  axis: int

  def __post_init__(self):
    if self.axis >= self.coordinate.ndim:
      raise ValueError(
          f'Dimension {self.axis=} of {self.coordinate=} is out of bounds'
      )

  @property
  def dims(self) -> tuple[AxisName, ...]:
    """Dimension names of the coordinate."""
    return (self.coordinate.dims[self.axis],)

  @property
  def shape(self) -> tuple[int, ...]:
    """Shape of the coordinate."""
    return (self.coordinate.shape[self.axis],)

  @property
  def fields(self) -> dict[AxisName, Field]:
    """A maps from field names to their values."""
    return self.coordinate.fields

  @property
  def ndim(self) -> int:
    """Dimensionality of the coordinate."""
    return 1

  def __repr__(self):
    return f'coordax.SelectedAxis({self.coordinate!r}, axis={self.axis})'


def consolidate_coordinates(*coordinates: Coordinate) -> tuple[Coordinate, ...]:
  """Consolidates coordinates without SelectedAxis objects, if possible."""
  axes = []
  result = []

  def reset_axes():
    result.extend(axes)
    axes[:] = []

  def append_axis(c):
    axes.append(c)
    if len(axes) == c.coordinate.ndim:
      # sucessful consolidation
      result.append(c.coordinate)
      axes[:] = []

  for c in coordinates:
    if isinstance(c, SelectedAxis) and c.axis == 0:
      # new SelectedAxis to consider consolidating
      reset_axes()
      append_axis(c)
    elif (
        isinstance(c, SelectedAxis)
        and axes
        and c.axis == len(axes)
        and c.coordinate == axes[-1].coordinate
    ):
      # continued SelectedAxis to consolidate
      append_axis(c)
    else:
      # coordinate cannot be consolidated
      reset_axes()
      result.append(c)

  reset_axes()

  return tuple(result)


@struct.pytree_dataclass
class CartesianProduct(Coordinate, struct.Struct):
  """Coordinate defined as the outer product of independent coordinates."""

  coordinates: tuple[Coordinate, ...] = dataclasses.field(
      metadata={'pytree_node': False}
  )

  def __post_init__(self):
    new_coordinates = []
    for c in self.coordinates:
      if isinstance(c, CartesianProduct):
        # c.coordinates are already Axiss where needed.
        new_coordinates.extend(c.coordinates)
      elif c.ndim > 1:
        for i in range(c.ndim):
          new_coordinates.append(SelectedAxis(c, i))
      else:
        new_coordinates.append(c)
    combined_coordinates = consolidate_coordinates(*new_coordinates)
    if len(combined_coordinates) <= 1:
      raise ValueError('CartesianProduct must contain more than 1 component')
    existing_dims = collections.Counter()
    for c in new_coordinates:
      existing_dims.update(c.dims)
    repeated_dims = [dim for dim, count in existing_dims.items() if count > 1]
    if repeated_dims:
      raise ValueError(f'CartesianProduct components contain {repeated_dims=}')
    object.__setattr__(self, 'coordinates', tuple(new_coordinates))

  def __eq__(self, other):
    if not isinstance(other, CartesianProduct):
      return len(self.coordinates) == 1 and self.coordinates[0] == other
    return isinstance(other, CartesianProduct) and all(
        self.coordinates[i] == other.coordinates[i]
        for i in range(len(self.coordinates))
    )

  @property
  def dims(self):
    return sum([c.dims for c in self.coordinates], start=tuple())

  @property
  def shape(self) -> tuple[int, ...]:
    """Returns the shape of the coordinate axes."""
    return sum([c.shape for c in self.coordinates], start=tuple())

  @property
  def fields(self) -> dict[AxisName, Field]:
    """Returns a mapping from field names to their values."""
    return functools.reduce(
        operator.or_, [c.fields for c in self.coordinates], {}
    )


@struct.pytree_dataclass
class NamedAxis(Coordinate, struct.Struct):
  """One dimensional coordinate that only dimension size."""

  name: AxisName = dataclasses.field(metadata={'pytree_node': False})
  size: int = dataclasses.field(metadata={'pytree_node': False})

  @property
  def dims(self) -> tuple[AxisName, ...]:
    return (self.name,)

  @property
  def shape(self) -> tuple[int, ...]:
    return (self.size,)

  @property
  def fields(self) -> dict[AxisName, Field]:
    return {}

  def __repr__(self):
    return f'coordax.NamedAxis({self.name!r}, size={self.size})'


# TODO(dkochkov): consider using @struct.pytree_dataclass here and storing
# tuple values instead of np.ndarray (which could be exposed as a property).
@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass
class LabeledAxis(Coordinate):  # pytype: disable=final-error
  """One dimensional coordinate with custom coordinate values."""

  name: AxisName = dataclasses.field(metadata={'pytree_node': False})
  ticks: np.ndarray = dataclasses.field(metadata={'pytree_node': False})

  @property
  def dims(self) -> tuple[AxisName, ...]:
    return (self.name,)

  @property
  def shape(self) -> tuple[int, ...]:
    return self.ticks.shape

  @property
  def fields(self) -> dict[AxisName, Field]:
    return {self.name: wrap(self.ticks, self)}

  def _components(self):
    return (self.name, ArrayKey(self.ticks))

  def tree_flatten(self):
    """Flattens LabeledAxis."""
    aux_data = self._components()
    return (), aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, leaves):
    """Unflattens LabeledAxis."""
    del leaves  # unused
    name, array_key = aux_data
    return cls(name=name, ticks=array_key.value)

  def __eq__(self, other):
    return (
        isinstance(other, LabeledAxis)
        and self._components() == other._components()
    )

  def __ne__(self, other):
    return not self == other

  def __hash__(self) -> int:
    return hash(self._components())

  def __repr__(self):
    return f'coordax.LabeledAxis({self.name!r}, ticks={self.ticks!r})'


def compose_coordinates(*coordinates: Coordinate) -> Coordinate:
  """Composes `coords` into a single coordinate system by cartesian product."""
  if not coordinates:
    raise ValueError('No coordinates provided.')
  coordinate_axes = []
  for c in coordinates:
    if isinstance(c, CartesianProduct):
      coordinate_axes.extend(c.coordinates)
    else:
      coordinate_axes.append(c)
  coordinates = consolidate_coordinates(*coordinate_axes)
  if len(coordinates) == 1:
    return coordinates[0]
  return CartesianProduct(coordinates)


def _dimension_names(*names: AxisName | Coordinate) -> tuple[AxisName, ...]:
  """Returns a tuple of dimension names from a list of names or coordinates."""
  dims_or_name_tuple = lambda x: x.dims if isinstance(x, Coordinate) else (x,)
  return sum([dims_or_name_tuple(c) for c in names], start=tuple())


def _validate_matching_coords(
    *axis_order: AxisName | Coordinate,
    coords: dict[AxisName, Coordinate],
):
  """Validates `axis_order` entries match with coordinates in `coords`."""
  coordinates = []
  for axis in axis_order:
    if isinstance(axis, Coordinate):
      if isinstance(axis, CartesianProduct):
        for c in axis.coordinates:
          coordinates.append(c)
      elif axis.ndim > 1:
        for i in range(axis.ndim):
          coordinates.append(SelectedAxis(axis, i))
      else:
        coordinates.append(axis)
  for c in coordinates:
    for dim in c.dims:
      if dim not in coords:
        raise ValueError(f'{dim=} of {c} not found in {coords.keys()=}')
      if coords[dim] != c:
        raise ValueError(f'Coordinate {c} does not match {coords[dim]=}')


def is_field(value) -> TypeGuard[Field]:
  """Returns True if `value` is of type `Field`."""
  return isinstance(value, Field)


def is_positional_prefix_field(f: Field) -> bool:
  """Returns True if positional axes of `f` are in prefix order."""
  return isinstance(f.named_array, named_axes.NamedArray)


def cmap(fun: Callable[..., Any]) -> Callable[..., Any]:
  """Vectorizes `fun` over coordinate dimensions of ``Field`` inputs.

  Args:
    fun: Function to vectorize over coordinate dimensions.

  Returns:
    A vectorized version of `fun` that applies original `fun` to locally
    positional dimensions in inputs, while vectorizing over all coordinate
    dimensions. All dimensions over which `fun` is vectorized will be present in
    every output.
  """
  if hasattr(fun, '__name__'):
    fun_name = fun.__name__
  else:
    fun_name = repr(fun)
  if hasattr(fun, '__doc__'):
    fun_doc = fun.__doc__
  else:
    fun_doc = None
  return _cmap_with_doc(fun, fun_name, fun_doc)


def _cmap_with_doc(
    fun: Callable[..., Any], fun_name: str, fun_doc: str | None = None
) -> Callable[..., Any]:
  """Builds a coordinate-vectorized wrapped function with a docstring."""

  @functools.wraps(fun)
  def wrapped_fun(*args, **kwargs):
    leaves, treedef = jax.tree.flatten((args, kwargs), is_leaf=is_field)
    field_leaves = [leaf for leaf in leaves if is_field(leaf)]
    all_coords = {}
    for field in field_leaves:
      for dim_name, c in field.coords.items():
        if dim_name in all_coords and all_coords[dim_name] != c:
          other = all_coords[dim_name]
          raise ValueError(f'Coordinates {c=} != {other=} use same {dim_name=}')
        else:
          all_coords[dim_name] = c
    named_array_leaves = [x.named_array if is_field(x) else x for x in leaves]
    fun_on_named_arrays = named_axes.nmap(fun)
    na_args, na_kwargs = jax.tree.unflatten(treedef, named_array_leaves)
    result = fun_on_named_arrays(*na_args, **na_kwargs)

    def _wrap_field(leaf):
      if isinstance(leaf, named_axes.NamedArray):
        return Field(
            named_array=leaf,
            coords={k: all_coords[k] for k in leaf.named_axes.keys()},
        )
      else:
        assert isinstance(leaf, named_axes.NamedArrayView)
        return Field(
            named_array=leaf,
            coords={k: all_coords[k] for k in leaf.data_axis_for_name.keys()},
        )

    return jax.tree.map(_wrap_field, result, is_leaf=named_axes.is_namedarray)

  docstr = (
      f'Dimension-vectorized version of `{fun_name}`. Takes similar arguments'
      f' as `{fun_name}` but accepts and returns Fields in place of arrays.'
  )
  if fun_doc:
    docstr += f'\n\nOriginal documentation:\n\n{fun_doc}'
  wrapped_fun.__doc__ = docstr
  return wrapped_fun


def _swapped_binop(binop):
  """Swaps the order of operations for a binary operation."""

  def swapped(x, y):
    return binop(y, x)

  return swapped


def _wrap_scalar_conversion(scalar_conversion):
  """Wraps a scalar conversion operator on a Field."""

  def wrapped_scalar_conversion(self: Field):
    if self.named_shape or self.positional_shape:
      raise ValueError(
          f'Cannot convert a non-scalar Field with {scalar_conversion}'
      )
    return scalar_conversion(self.unwrap())

  return wrapped_scalar_conversion


def _wrap_array_method(name):
  """Wraps an array method on a Field."""

  def func(array, *args, **kwargs):
    return getattr(array, name)(*args, **kwargs)

  array_method = getattr(jax.Array, name)
  wrapped_func = cmap(func)
  functools.update_wrapper(
      wrapped_func,
      array_method,
      assigned=('__name__', '__qualname__', '__annotations__'),
      updated=(),
  )
  wrapped_func.__module__ = __name__
  wrapped_func.__doc__ = (
      'Name-vectorized version of array method'
      f' `{name} <numpy.ndarray.{name}>`. Takes similar arguments as'
      f' `{name} <numpy.ndarray.{name}>` but accepts and returns Fields'
      ' (or FieldViews) in place of regular arrays.'
  )
  return wrapped_func


def _wrap_array_or_scalar(inputs: float | np.ndarray | jax.Array) -> Field:
  """Helper function that wraps inputs in a fully positional Field."""
  if isinstance(inputs, float):
    data_array = jnp.array(inputs)
  elif isinstance(inputs, np.ndarray):
    data_array = inputs
  else:
    data_array = jnp.asarray(inputs)
  named_array = named_axes.NamedArray(
      named_axes=collections.OrderedDict(), data_array=data_array
  )
  wrapped = Field(
      named_array=named_array,
      coords={},
  )
  return wrapped


@struct.pytree_dataclass
class Field(struct.Struct):
  """An array with a combination of positional and named dimensions.

  Attributes:
    named_array: A named array that tracks positional and named dimensions.
    coords: A mapping from dimension names to their coordinate objects.
  """

  # TODO(shoyer): Consider storing an ndarray and dims separately, and creating
  # the Penzai NamedArray on the fly. This would be make Penzai more of an
  # implementation detail instead of part of the public API.
  # Potential downside: would need a separate wrapper for NamedArrayView.
  # e.g.,
  #   data: jax.Array
  #   dims: tuple[AxisName, ...]
  #   coord: dict[AxisName, Coordinate]
  # or
  #   data: jax.Array
  #   coords: tuple[Coordinate, ...]

  named_array: named_axes.NamedArray | named_axes.NamedArrayView
  coords: dict[AxisName, Coordinate] = dataclasses.field(
      metadata={'pytree_node': False}
  )

  @property
  def data(self) -> np.ndarray | jax.Array:
    """The value of the underlying data array."""
    return self.named_array.data_array

  @property
  def dtype(self) -> np.dtype:
    """The dtype of the field."""
    return self.named_array.dtype

  def check_valid(self) -> None:
    """Checks that the field coordinates and dimension names are consistent."""
    data_dims = set(self.named_array.named_shape.keys())
    keys_dims = set(self.coords.keys())
    coord_dims = set(sum([c.dims for c in self.coords.values()], start=()))
    if (data_dims != keys_dims) or (data_dims != coord_dims):
      raise ValueError(
          'Field dimension names must be the same across keys and coordinates'
          f' , got {data_dims=}, {keys_dims=} and {keys_dims=}.'
      )
    self.named_array.check_valid()

  @property
  def named_shape(self) -> Mapping[AxisName, int]:
    """A mapping of axis names to their sizes."""
    return self.named_array.named_shape

  @property
  def positional_shape(self) -> tuple[int, ...]:
    """A tuple of axis sizes for any anonymous axes."""
    return self.named_array.positional_shape

  @property
  def shape(self) -> tuple[int, ...]:
    """A tuple of axis sizes of the underlying data array."""
    return self.named_array.data_array.shape

  @property
  def dims(self) -> tuple[AxisName | int, ...]:
    """A tuple of indices and dimension names in the data layout order."""
    if isinstance(self.named_array, named_axes.NamedArray):
      postional_dims = tuple(range(len(self.positional_shape)))
      return postional_dims + tuple(self.named_array.named_axes.keys())
    elif isinstance(self.named_array, named_axes.NamedArrayView):
      data_axis_for_logical_axis = self.named_array.data_axis_for_logical_axis
      to_name = {v: k for k, v in self.named_array.data_axis_for_name.items()}
      to_logical = {v: k for k, v in enumerate(data_axis_for_logical_axis)}
      axis_to_dim = to_logical | to_name
      return tuple(
          axis_to_dim[i] for i in range(len(self.named_array.data_shape))
      )
    else:
      raise TypeError(
          f'{type(self.named_array)=} is not of expected NamedArray type.'
      )

  @property
  def coord_fields(self) -> dict[AxisName, Field]:
    """A mapping from coordinate field names to their values."""
    return functools.reduce(
        operator.or_, [c.fields for c in self.coords.values()], {}
    )

  def unwrap(self, *names: AxisName | Coordinate) -> jax.Array:
    """Extracts the underlying data from fully positional or named field."""
    return self.named_array.unwrap(*_dimension_names(*names))

  def with_positional_prefix(self) -> Field:
    """Returns a `Field` with positional axes moved to the front."""
    return Field(
        named_array=self.named_array.with_positional_prefix(),
        coords=self.coords,
    )

  def as_view(self) -> Field:
    """Returns a `Field` with a view representation of the underlying data."""
    return Field(
        named_array=self.named_array.as_namedarrayview(),
        coords=self.coords,
    )

  def untag(self, *axis_order: AxisName | Coordinate) -> Field:
    """Returns a view of the field with the requested axes made positional."""
    _validate_matching_coords(*axis_order, coords=self.coords)
    untag_dims = _dimension_names(*axis_order)
    result = Field(
        named_array=self.named_array.untag(*untag_dims),
        coords={k: v for k, v in self.coords.items() if k not in untag_dims},
    )
    result.check_valid()
    return result

  def tag(self, *names: AxisName | Coordinate) -> Field:
    """Returns a Field with attached coordinates to the positional axes."""
    tag_dims = _dimension_names(*names)
    tagged_array = self.named_array.tag(*tag_dims)
    coords = {}
    coords.update(self.coords)
    for c in names:
      if isinstance(c, Coordinate):
        if isinstance(c, CartesianProduct):
          for sub_c in c.coordinates:
            coords[sub_c.dims[0]] = sub_c
        else:
          for i, dim in enumerate(c.dims):
            coords[dim] = c if c.ndim == 1 else SelectedAxis(c, i)
      else:
        coords[c] = NamedAxis(c, size=tagged_array.named_shape[c])
    result = Field(named_array=tagged_array, coords=coords)
    result.check_valid()
    return result

  @property
  def ndim(self) -> int:
    return len(self.dims)

  def untag_prefix(self, *axis_order: AxisName | Coordinate) -> Field:
    """Returns a field with requested axes made front positional axes."""
    _validate_matching_coords(*axis_order, coords=self.coords)
    untag_dims = _dimension_names(*axis_order)
    result = Field(
        named_array=self.named_array.untag_prefix(*untag_dims),
        coords={k: v for k, v in self.coords.items() if k not in untag_dims},
    )
    result.check_valid()
    return result

  def tag_prefix(self, *axis_order: AxisName | Coordinate) -> Field:
    """Returns a field with coords attached to the first positional axes."""
    tag_dims = _dimension_names(*axis_order)
    tagged_array = self.named_array.tag_prefix(*tag_dims)
    coords = {}
    coords.update(self.coords)
    for c in axis_order:
      if isinstance(c, Coordinate):
        for dim in c.dims:
          coords[dim] = c
      else:
        coords[c] = NamedAxis(c, size=tagged_array.named_shape[c])
    result = Field(named_array=tagged_array, coords=coords)
    result.check_valid()
    return result

  def tag_suffix(self, *axis_order: AxisName | Coordinate) -> Field:
    """Returns a field with coords attached to the last positional axes."""
    n_tmp = len(self.positional_shape) - len(_dimension_names(*axis_order))
    tmp_dims = [named_axes.TmpPosAxisMarker() for _ in range(n_tmp)]
    return self.tag(*tmp_dims, *axis_order).untag_prefix(*tmp_dims)

  def untag_suffix(self, *axis_order: AxisName | Coordinate) -> Field:
    """Returns a field with requested axes made last positional axes."""
    n_tmp = len(self.positional_shape)
    tmp_dims = [named_axes.TmpPosAxisMarker() for _ in range(n_tmp)]
    return self.tag(*tmp_dims).untag(*tmp_dims, *axis_order)

  # Note: Can't call this "transpose" like Xarray, to avoid conflicting with the
  # positional only ndarray method.
  def order_as(self, *axis_order: AxisName | Coordinate) -> Field:
    """Returns a field with the axes in the given order."""
    _validate_matching_coords(*axis_order, coords=self.coords)
    ordered_dims = _dimension_names(*axis_order)
    ordered_array = self.named_array.order_as(*ordered_dims)
    result = Field(named_array=ordered_array, coords=self.coords)
    result.check_valid()
    return result

  def order_like(self, other: Field) -> Field:
    """Returns a field with the axes in the same order as `other`."""
    self.check_valid()
    other.check_valid()
    # To be able to order-like coordinates of two fields must match.
    _validate_matching_coords(*other.coords.values(), coords=self.coords)
    return Field(
        named_array=self.named_array.order_like(other.named_array),
        coords=self.coords,
    )

  def broadcast_to(
      self,
      positional_shape: Sequence[int] = (),
      named_shape: Mapping[AxisName, int] | tuple[Coordinate, ...] = (),
  ) -> Field:
    """Returns a field with positional and named shapes broadcasted."""
    if isinstance(named_shape, tuple):
      # TODO(dkochkov): Do we need to support broadcast_to with new coordinates?
      _validate_matching_coords(*named_shape, coords=self.coords)
      coordinates = [c for c in named_shape if isinstance(c, Coordinate)]
      dims = _dimension_names(*coordinates)
      shapes = sum([c.shape for c in coordinates], start=tuple())
      named_shape = {dim: size for dim, size in zip(dims, shapes)}
    return Field(
        named_array=self.named_array.broadcast_to(
            positional_shape, named_shape
        ),
        coords=self.coords,
    )

  def broadcast_like(self, other: Field | jax.typing.ArrayLike) -> Field:
    """Returns a field broadcasted to the shape of `other`."""
    if isinstance(other, Field):
      return self.broadcast_to(other.positional_shape, other.named_shape)
    else:
      shape = jnp.shape(other)
      return self.broadcast_to(shape, {})

  def __getitem__(self, indexer) -> Field:
    """Retrieves slices from an indexer, as in pz.core.named_axes indexing."""
    self.check_valid()
    # TODO(shoyer): consider disabling positional indexing, because it's
    # ambiguous whether it refers to positional-only or all axes.
    return Field(
        named_array=self.named_array.at[indexer].get(), coords=self.coords
    )

  # Iteration. Note that we *must* implement this to avoid Python simply trying
  # to run __getitem__ until it raises IndexError, because we won't raise
  # IndexError (since JAX clips array indices).
  def __iter__(self):
    if not self.positional_shape:
      raise ValueError('Cannot iterate over an array with no positional axes.')
    for i in range(self.positional_shape[0]):
      yield self[i]

  @classmethod
  def wrap(
      cls, array: jax.typing.ArrayLike | float, *names: AxisName | Coordinate
  ) -> Field:
    """Wraps a positional array as a ``Field``."""
    wrapped = _wrap_array_or_scalar(array)
    if names:
      return wrapped.tag(*names)
    else:
      return wrapped

  # TODO(shoyer): after updating the data model, the __repr__ should look like
  # either:
  #
  # Field(
  #     data=array(...),
  #     coords=(
  #         NamedAxis("time"),
  #         LatLonGrid(...),
  #     ),
  # )
  #
  # or:
  #
  # Field(
  #     data=array(...),
  #     dims=('time', 'lon', 'lat'),
  #     coords={
  #         'time': NamedAxis("time"),
  #         'lon': SelectedAxis(LonLatGrid(...), axis=0),
  #         'lat': SelectedAxis(LonLatGrid(...), axis=1),
  #     }
  # )

  # TODO(shoyer): restore a custom repr, once it's clear that we handle axis
  # order consistently with penzai.
  # def __repr__(self):
  #   indent = '    '
  #   data_repr = textwrap.indent(repr(self.data), prefix=indent)
  #   coordinates = consolidate_coordinates(*self.coords.values())
  #   coords_repr = indent + f',\n{indent}'.join(
  #       repr(c).removeprefix('coordax.') for c in coordinates
  #   )
  #   return f'coordax.Field.wrap(\n{data_repr},\n{coords_repr},\n)'

  # Convenience wrappers: Elementwise infix operators.
  __lt__ = _cmap_with_doc(operator.lt, 'jax.Array.__lt__')
  __le__ = _cmap_with_doc(operator.le, 'jax.Array.__le__')
  __eq__ = _cmap_with_doc(operator.eq, 'jax.Array.__eq__')
  __ne__ = _cmap_with_doc(operator.ne, 'jax.Array.__ne__')
  __ge__ = _cmap_with_doc(operator.ge, 'jax.Array.__ge__')
  __gt__ = _cmap_with_doc(operator.gt, 'jax.Array.__gt__')

  __add__ = _cmap_with_doc(operator.add, 'jax.Array.__add__')
  __sub__ = _cmap_with_doc(operator.sub, 'jax.Array.__sub__')
  __mul__ = _cmap_with_doc(operator.mul, 'jax.Array.__mul__')
  __truediv__ = _cmap_with_doc(operator.truediv, 'jax.Array.__truediv__')
  __floordiv__ = _cmap_with_doc(operator.floordiv, 'jax.Array.__floordiv__')
  __mod__ = _cmap_with_doc(operator.mod, 'jax.Array.__mod__')
  __divmod__ = _cmap_with_doc(divmod, 'jax.Array.__divmod__')
  __pow__ = _cmap_with_doc(operator.pow, 'jax.Array.__pow__')
  __lshift__ = _cmap_with_doc(operator.lshift, 'jax.Array.__lshift__')
  __rshift__ = _cmap_with_doc(operator.rshift, 'jax.Array.__rshift__')
  __and__ = _cmap_with_doc(operator.and_, 'jax.Array.__and__')
  __or__ = _cmap_with_doc(operator.or_, 'jax.Array.__or__')
  __xor__ = _cmap_with_doc(operator.xor, 'jax.Array.__xor__')

  __radd__ = _cmap_with_doc(_swapped_binop(operator.add), 'jax.Array.__radd__')
  __rsub__ = _cmap_with_doc(_swapped_binop(operator.sub), 'jax.Array.__rsub__')
  __rmul__ = _cmap_with_doc(_swapped_binop(operator.mul), 'jax.Array.__rmul__')
  __rtruediv__ = _cmap_with_doc(
      _swapped_binop(operator.truediv), 'jax.Array.__rtruediv__'
  )
  __rfloordiv__ = _cmap_with_doc(
      _swapped_binop(operator.floordiv), 'jax.Array.__rfloordiv__'
  )
  __rmod__ = _cmap_with_doc(_swapped_binop(operator.mod), 'jax.Array.__rmod__')
  __rdivmod__ = _cmap_with_doc(_swapped_binop(divmod), 'jax.Array.__rdivmod__')
  __rpow__ = _cmap_with_doc(_swapped_binop(operator.pow), 'jax.Array.__rpow__')
  __rlshift__ = _cmap_with_doc(
      _swapped_binop(operator.lshift), 'jax.Array.__rlshift__'
  )
  __rrshift__ = _cmap_with_doc(
      _swapped_binop(operator.rshift), 'jax.Array.__rrshift__'
  )
  __rand__ = _cmap_with_doc(_swapped_binop(operator.and_), 'jax.Array.__rand__')
  __ror__ = _cmap_with_doc(_swapped_binop(operator.or_), 'jax.Array.__ror__')
  __rxor__ = _cmap_with_doc(_swapped_binop(operator.xor), 'jax.Array.__rxor__')

  __abs__ = _cmap_with_doc(operator.abs, 'jax.Array.__abs__')
  __neg__ = _cmap_with_doc(operator.neg, 'jax.Array.__neg__')
  __pos__ = _cmap_with_doc(operator.pos, 'jax.Array.__pos__')
  __invert__ = _cmap_with_doc(operator.inv, 'jax.Array.__invert__')

  # Convenience wrappers: Scalar conversions.
  __bool__ = _wrap_scalar_conversion(bool)
  __complex__ = _wrap_scalar_conversion(complex)
  __int__ = _wrap_scalar_conversion(int)
  __float__ = _wrap_scalar_conversion(float)
  __index__ = _wrap_scalar_conversion(operator.index)

  # elementwise operations
  astype = _wrap_array_method('astype')
  clip = _wrap_array_method('clip')
  conj = _wrap_array_method('conj')
  conjugate = _wrap_array_method('conjugate')
  imag = _wrap_array_method('imag')
  real = _wrap_array_method('real')
  round = _wrap_array_method('round')
  view = _wrap_array_method('view')

  # Intentionally not included: anything that acts on a subset of axes or takes
  # an axis as an argument (e.g., mean). It is ambiguous whether these should
  # act over positional or named axes.

  # maybe include some of below with names that signify positional nature?
  # reshape = _wrap_array_method('reshape')
  # squeeze = _wrap_array_method('squeeze')
  # transpose = _wrap_array_method('transpose')
  # T = _wrap_array_method('T')
  # mT = _wrap_array_method('mT')  # pylint: disable=invalid-name


wrap = Field.wrap


def wrap_like(array: jax.Array, other: Field) -> Field:
  """Wraps `array` with the same coordinates as `other`."""
  other.check_valid()
  if array.shape != other.shape:
    raise ValueError(f'{array.shape=} and {other.shape=} must be equal')
  return Field(
      named_array=dataclasses.replace(other.named_array, data_array=array),
      coords=other.coords,
  )



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for converting between xarray and DataObservation objects."""

from typing import Hashable

from dinosaur import spherical_harmonic
from dinosaur import xarray_utils as dino_xarray_utils
import jax.numpy as jnp
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import data_specs
from neuralgcm.experimental import scales
from neuralgcm.experimental import typing
from neuralgcm.experimental import units
import numpy as np
import xarray


AxisName = Hashable
XR_SHORT_LON_NAME = 'lon'
XR_SHORT_LAT_NAME = 'lat'
XR_LON_NAME = 'longitude'
XR_LAT_NAME = 'latitude'


def _wrap_suffix(array: typing.Array, *names: AxisName | cx.Coordinate):
  return cx.wrap(array).tag_suffix(*names).with_positional_prefix()


verify_grid_consistency = dino_xarray_utils.verify_grid_consistency


def xarray_nondimensionalize(
    ds: xarray.Dataset | xarray.DataArray,
    sim_units: units.SimUnits,
) -> xarray.Dataset:
  return xarray.apply_ufunc(sim_units.nondimensionalize, ds)


def coordinates_from_dataset(
    ds: xarray.Dataset,
    spherical_harmonics_impl=spherical_harmonic.FastSphericalHarmonics,
) -> coordinates.DinosaurCoordinates:
  """Infers coordinates from `ds`."""
  # TODO(dkochkov): Generalize this to take a role of composing coordinates
  # from a collection of candidate coordax.Coordinate instances.
  dino_coords = dino_xarray_utils.coordinate_system_from_dataset(
      ds,
      truncation=dino_xarray_utils.LINEAR,
      spherical_harmonics_impl=spherical_harmonics_impl,
  )
  return coordinates.DinosaurCoordinates.from_dinosaur_coords(dino_coords)


def get_longitude_latitude_names(ds: xarray.Dataset) -> tuple[str, str]:
  """Infers names used for longitude and latitude in the dataset `ds`."""
  if XR_SHORT_LON_NAME in ds.dims and XR_SHORT_LAT_NAME in ds.dims:
    return (XR_SHORT_LON_NAME, XR_SHORT_LAT_NAME)
  if XR_LON_NAME in ds.dims and XR_LAT_NAME in ds.dims:
    return (XR_LON_NAME, XR_LAT_NAME)
  raise ValueError(f'No `lon/lat`|`longitude/latitude` in {ds.coords.keys()=}')


def attach_data_units(
    array: xarray.DataArray,
    default_units: typing.Quantity = typing.units.dimensionless,
) -> xarray.DataArray:
  """Attaches units to `array` based on `attrs.units` or `default_units`."""
  attrs = dict(array.attrs)
  unit = attrs.pop('units', None)
  if unit is not None:
    data = units.parse_units(unit) * array.data
  else:
    data = default_units * array.data
  return xarray.DataArray(data, array.coords, array.dims, attrs=attrs)


def nodal_orography_from_ds(ds: xarray.Dataset) -> xarray.DataArray:
  """Returns orography in nodal representation from `ds`."""
  orography_key = dino_xarray_utils.OROGRAPHY
  if orography_key not in ds:
    ds[orography_key] = (
        ds[dino_xarray_utils.GEOPOTENTIAL_AT_SURFACE_KEY]
        / scales.GRAVITY_ACCELERATION.magnitude
    )
  lon_lat_order = get_longitude_latitude_names(ds)
  orography = attach_data_units(ds[orography_key], typing.units.meter)
  return orography.transpose(*lon_lat_order)


def _xarray_to_data_dict(dataset, *, values: str = 'values'):
  """Extracts `values` from `dataset` and returns as a dictionary."""
  expected_dims_order = (
      dino_xarray_utils.XR_TIME_NAME,
      dino_xarray_utils.XR_LEVEL_NAME,
      dino_xarray_utils.XR_LON_NAME,
      dino_xarray_utils.XR_LAT_NAME,
  )
  dims = []
  for dim in expected_dims_order:
    if dim in dataset.dims:
      dims.append(dim)

  dataset = dataset.transpose(..., *dims)
  data = {}
  for k in dataset:
    assert isinstance(k, str)  # satisfy pytype
    v = getattr(dataset[k], values)
    data[k] = v
  return data


def xarray_to_data_dict(ds: xarray.Dataset, *, values: str = 'values'):
  lon_name, lat_name = get_longitude_latitude_names(ds)
  if lon_name != XR_SHORT_LON_NAME:
    ds = ds.rename({lon_name: XR_SHORT_LON_NAME})
  if lat_name != XR_SHORT_LAT_NAME:
    ds = ds.rename({lat_name: XR_SHORT_LAT_NAME})
  return _xarray_to_data_dict(ds, values=values)


# TODO(dkochkov): Transition to using typing.Timedelta instead of sim_time.


def datetime64_to_nondim_time(
    time: np.ndarray,
    sim_units: units.SimUnits,
) -> np.ndarray:
  """Converts `time` in datetime64 format to nondimensional sim_time."""
  return np.asarray(
      sim_units.nondimensionalize(
          ((time - sim_units.reference_datetime) / np.timedelta64(1, 'h'))
          * typing.units.hour
      )
  )


def nondim_time_delta_from_time_axis(
    time: typing.Array,
    sim_units: units.SimUnits,
) -> typing.Numeric:
  """Infers time delta along `time` axis in nondimensional units."""
  time_delta = time[1] - time[0]
  if not np.issubdtype(time.dtype, np.floating):
    time_delta = np.timedelta64(time_delta, 's') / np.timedelta64(1, 's')
    result = sim_units.nondimensionalize(time_delta * typing.units.second)
    assert isinstance(result, typing.Numeric)
    return result
  return float(time_delta)


def with_sim_time(ds: xarray.Dataset, sim_units: units.SimUnits):
  """Returns `ds` with nondimensional time added as `sim_time` if absent."""
  if 'sim_time' in ds:
    return ds
  if np.issubdtype(ds.time.dtype, np.floating):
    nondim_time = ds.time.data
  else:
    nondim_time = datetime64_to_nondim_time(ds.time.data, sim_units)
  # if dataset contains `sample` axis, sim_time should have it as well.
  if dino_xarray_utils.XR_SAMPLE_NAME in ds.coords:
    nondim_time = nondim_time[np.newaxis, ...]
    nondim_time = np.repeat(
        nondim_time, ds.sizes[dino_xarray_utils.XR_SAMPLE_NAME], 0
    )
    sim_time = ((dino_xarray_utils.XR_SAMPLE_NAME,) + ds.time.dims, nondim_time)
  else:
    sim_time = (ds.time.dims, nondim_time)
  return ds.assign(sim_time=sim_time)


def xarray_to_timed_fields_dict(
    ds: xarray.Dataset,
    *,
    values: str = 'values',
    coords: cx.Coordinate | None = None,
):
  """Extracts data from `ds` to a dictionary of `StaticField`s."""
  if coords is None:
    coords = coordinates_from_dataset(ds)
  data_dict = xarray_to_data_dict(ds, values=values)
  if 'sim_time' not in data_dict:
    raise ValueError(f'Dataset {ds.coords.keys()} does not contain sim_time.')
  sim_time = data_dict.pop('sim_time')
  observations = {
      k: data_specs.TimedField(_wrap_suffix(v, coords), sim_time)
      for k, v in data_dict.items()
  }
  return observations


def xarray_to_timed_fieldset(
    ds: xarray.Dataset,
    *,
    values: str = 'values',
    coords: cx.Coordinate | None = None,
):
  """Extracts data from `ds` to a dictionary of `StaticField`s."""
  if coords is None:
    coords = coordinates_from_dataset(ds)  # use default unnamed dims for now.
  data_dict = xarray_to_data_dict(ds, values=values)
  if 'sim_time' not in data_dict:
    raise ValueError(f'Dataset {ds.coords.keys()} does not contain sim_time.')
  sim_time = data_dict.pop('sim_time')
  return data_specs.TimedObservations(
      fields={k: _wrap_suffix(v, coords) for k, v in data_dict.items()},
      timestamp=sim_time,
  )


def timed_field_to_xarray(
    fields_dict: dict[str, data_specs.TimedField[cx.Field]],
    *,
    sim_units: units.SimUnits | None = None,
    additional_coords: dict[str, typing.Array] | None = None,
    serialize_coords_to_attrs: bool = False,
):
  """Converts `fields_dict` to an xarray dataset."""
  data_dict = {k: v.field.data for k, v in fields_dict.items()}
  sample_obs = next(iter(fields_dict.values()))
  sim_time = np.asarray(sample_obs.timestamp)
  level_options = set(['sigma', 'pressure', 'level', 'layer_index'])
  data_dims = sample_obs.field.dims
  level_name = level_options.intersection(set(data_dims))
  if len(level_name) != 1:
    raise ValueError(f'Need exactly 1 match {level_options=} & {data_dims=}')
  (level_name,) = level_name
  vertical = sample_obs.field.coords[level_name]
  horizontal_options = set(
      ['longitude', 'latitude', 'longitudinal_wavenumber', 'total_wavenumber']
  )
  horizontal_names = horizontal_options.intersection(set(data_dims))
  if len(horizontal_names) != 2:
    raise ValueError(f'Expected 2 matches {horizontal_options=} & {data_dims=}')
  horizontal = cx.compose_coordinates(
      *[sample_obs.field.coords[k] for k in horizontal_names]
  )
  if (
      sim_units is not None
      and sim_time is not None
      and jnp.isdtype(sim_time.dtype, kind='real floating')
  ):
    sim_time = sim_units.sim_time_to_datetime64(sim_time)

  dino_coords = coordinates.DinosaurCoordinates(
      horizontal=horizontal, vertical=vertical
  )
  return dino_xarray_utils.data_to_xarray(
      data={k: v for k, v in data_dict.items()},
      coords=dino_coords.dinosaur_coords,
      times=sim_time,
      sample_ids=None,
      additional_coords=additional_coords,
      serialize_coords_to_attrs=serialize_coords_to_attrs,
  )



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modules that interpolate data from one coordinate system to another."""

import dataclasses

import jax
import jax.numpy as jnp
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import typing
import numpy as np


@dataclasses.dataclass(frozen=True)
class SpectralRegridder:
  """Regrid between spherical harmonic grids with truncation or zero padding."""

  target_coords: coordinates.SphericalHarmonicGrid

  def truncate_to_target_wavenumbers(
      self,
      field: cx.Field,
  ) -> cx.Field:
    """Interpolates to lower resolution spherical harmonic by truncation."""
    # TODO(dkochkov) consider using coordinate values to inform truncation.
    target_lon_wavenumbers = self.target_coords.sizes['longitude_wavenumber']
    target_total_wavenumbers = self.target_coords.sizes['total_wavenumber']
    lon_slice = slice(0, target_lon_wavenumbers)
    total_slice = slice(0, target_total_wavenumbers)
    field = field.untag_prefix('longitude_wavenumber', 'total_wavenumber')
    result = cx.cmap(lambda x: x[lon_slice, total_slice])(field)
    return result.tag(self.target_coords)

  def pad_to_target_wavenumbers(
      self,
      field: cx.Field,
  ) -> cx.Field:
    """Interpolates to higher resolution spherical harmonic by zero-padding."""
    # TODO(dkochkov) use `sizes` on coords to carry shape of dims info.
    input_lon_k = field.coord_fields['longitude_wavenumber'].shape[0]
    input_total_k = field.coord_fields['total_wavenumber'].shape[0]
    target_lon_k = self.target_coords.sizes['longitude_wavenumber']
    target_total_k = self.target_coords.sizes['total_wavenumber']
    pad_lon = (0, target_lon_k - input_lon_k)
    pad_total = (0, target_total_k - input_total_k)
    pad_fn = lambda x: jnp.pad(x, pad_width=(pad_lon, pad_total))
    field = field.untag_prefix('longitude_wavenumber', 'total_wavenumber')
    result = cx.cmap(pad_fn)(field)
    return result.tag_prefix(self.target_coords)

  def interpolate_field(self, field: cx.Field) -> cx.Field:
    """Interpolates a single field."""
    # TODO(dkochkov) Check that inputs.coords includes SphericalHarmonicGrid.
    input_lon_k = field.coord_fields['longitude_wavenumber'].shape[0]
    input_total_k = field.coord_fields['total_wavenumber'].shape[0]
    target_lon_k = self.target_coords.sizes['longitude_wavenumber']
    target_total_k = self.target_coords.sizes['total_wavenumber']
    if (input_total_k < target_total_k) and (input_lon_k < target_lon_k):
      return self.pad_to_target_wavenumbers(field)
    elif (input_total_k >= target_total_k) and (input_lon_k >= target_lon_k):
      return self.truncate_to_target_wavenumbers(field)
    else:
      raise ValueError(
          'Incompatible horizontal coordinates with shapes '
          f'{field.dims=} with {field.shape=}, '
          f'{self.target_coords.dims=} with {self.target_coords.shape=}, '
      )

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    """Interpolates fields in `inputs`."""
    is_field = lambda x: isinstance(x, cx.Field)
    return jax.tree.map(
        self.interpolate_field,
        inputs,
        is_leaf=is_field,
    )



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modules that parameterize unsimulated processes."""

from typing import Callable

from flax import nnx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import pytree_mappings
from neuralgcm.experimental import pytree_transforms
from neuralgcm.experimental import pytree_utils
from neuralgcm.experimental import spatial_filters
from neuralgcm.experimental import typing


ShapeFloatStruct = typing.ShapeFloatStruct


class ModalNeuralDivCurlParameterization(nnx.Module):
  """Computes modal tendencies with `u, v` → `δ, ζ` transform."""

  def __init__(
      self,
      *,
      coords: coordinates.DinosaurCoordinates,
      surface_field_names: tuple[str, ...],
      volume_field_names: tuple[str, ...],
      features_module: pytree_transforms.Transform,
      mapping_factory: Callable[
          ..., pytree_mappings.ChannelMapping | pytree_mappings.VariableMapping
      ],
      tendency_transform: pytree_transforms.Transform,
      modal_filter: spatial_filters.ModalSpatialFilter,
      input_state_shapes: typing.Pytree | None = None,
      u_key: str = 'u_component_of_wind',
      v_key: str = 'v_component_of_wind',
      rngs: nnx.Rngs,
  ):
    output_shapes = {}
    # TODO(dkochkov): Add checks for the field names that are required.
    uv_fields = set([u_key, v_key])
    div_curl_fields = set(['divergence', 'vorticity'])
    if len(div_curl_fields.intersection(volume_field_names)) != 2:
      raise ValueError('Volume fields must contain `divergence & vorticity`.')
    layers = coords.vertical.shape[0]
    dinosaur_grid = coords.dinosaur_grid
    # TODO(dkochkov): Compute these using coords modifications.
    for name in (set(volume_field_names) | uv_fields) - div_curl_fields:
      output_shapes[name] = ShapeFloatStruct(
          (layers,) + dinosaur_grid.nodal_shape
      )
    for name in set(surface_field_names):
      output_shapes[name] = ShapeFloatStruct((1,) + dinosaur_grid.nodal_shape)
    if input_state_shapes is None:
      input_state_shapes = pytree_mappings.MINIMAL_STATE_STRUCT  # default.
    input_shapes = features_module.output_shapes(input_state_shapes)
    self.parameterization_mapping = mapping_factory(
        input_shapes=input_shapes,
        output_shapes=output_shapes,
        rngs=rngs,
    )
    self.features_module = features_module
    self.tendency_transform = tendency_transform
    self.to_div_curl = pytree_transforms.ToModalWithDivCurl(coords.horizontal)
    self.filter = modal_filter

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    # inputs = self.coords.with_dycore_sharding(inputs)
    inputs_dict, from_dict_fn = pytree_utils.as_dict(inputs)
    features = self.features_module(inputs_dict)
    # features = self.coords.dycore_to_physics_sharding(features)
    tendencies = self.parameterization_mapping(features)
    # tendencies = self.coords.physics_to_dycore_sharding(tendencies)
    tendencies = self.tendency_transform(tendencies)
    modal_tendencies = self.to_div_curl(tendencies)
    modal_tendencies = self.filter.filter_modal(modal_tendencies)
    modal_tendencies = pytree_utils.replace_with_matching_or_default(
        inputs_dict, modal_tendencies, default=0.0
    )
    outputs = from_dict_fn(modal_tendencies)
    # outputs = self.coords.with_dycore_sharding(outputs)
    return outputs



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modules parameterizing PDEs that describe dynamical or physical processes."""

import dataclasses
from typing import Callable, Sequence

from dinosaur import primitive_equations
from dinosaur import sigma_coordinates
from flax import nnx
import jax
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import orographies
from neuralgcm.experimental import pytree_mappings
from neuralgcm.experimental import pytree_transforms
from neuralgcm.experimental import pytree_utils
from neuralgcm.experimental import time_integrators
from neuralgcm.experimental import typing
from neuralgcm.experimental import units
import numpy as np


ImplicitExplicitODE = time_integrators.ImplicitExplicitODE
ExplicitODE = time_integrators.ExplicitODE
ShapeFloatStruct = typing.ShapeFloatStruct


class PrimitiveEquations(ImplicitExplicitODE):
  """Equation module for moist primitive equations ."""

  def __init__(
      self,
      coords: coordinates.DinosaurCoordinates,
      sim_units: units.SimUnits,
      reference_temperatures: Sequence[float],
      orography_module: orographies.ModalOrography,
      vertical_advection: Callable[..., typing.Array] = (
          sigma_coordinates.centered_vertical_advection
      ),
      equation_cls=primitive_equations.MoistPrimitiveEquationsWithCloudMoisture,
      include_vertical_advection: bool = True,
  ):
    self.coords = coords
    self.orography_module = orography_module
    self.sim_units = sim_units
    self.orography = orography_module
    self.reference_temperatures = reference_temperatures
    self.vertical_advection = vertical_advection
    self.include_vertical_advection = include_vertical_advection
    self.equation_cls = equation_cls

  @property
  def primitive_equation(self):
    return self.equation_cls(
        coords=self.coords.dinosaur_coords,
        physics_specs=self.sim_units,
        reference_temperature=np.asarray(self.reference_temperatures),
        orography=self.orography_module.modal_orography,
        vertical_advection=self.vertical_advection,
        include_vertical_advection=self.include_vertical_advection,
    )

  @property
  def T_ref(self) -> typing.Array:
    return self.primitive_equation.T_ref

  def explicit_terms(
      self, state: primitive_equations.StateWithTime
  ) -> primitive_equations.StateWithTime:
    return self.primitive_equation.explicit_terms(state)

  def implicit_terms(
      self, state: primitive_equations.StateWithTime
  ) -> primitive_equations.StateWithTime:
    return self.primitive_equation.implicit_terms(state)

  def implicit_inverse(
      self, state: primitive_equations.StateWithTime, step_size: float
  ) -> primitive_equations.StateWithTime:
    return self.primitive_equation.implicit_inverse(state, step_size)


class SimTimeEquation(time_integrators.ExplicitODE):
  """Equation module describing evolution of `sim_time` variable."""

  def explicit_terms(self, x: typing.Pytree) -> typing.Pytree:
    x_dict, from_dict_fn = pytree_utils.as_dict(x)
    if 'sim_time' not in x_dict:
      raise ValueError(f'sim_time not found in {x_dict.keys()}')
    terms = {k: None if k != 'sim_time' else 1.0 for k in x_dict.keys()}
    return from_dict_fn(terms)


class NeuralEquation(time_integrators.ExplicitODE):
  """Module for equations fully parameterized by a neural network."""

  def __init__(
      self,
      *,
      coords: coordinates.DinosaurCoordinates,
      surface_field_names: tuple[str, ...],
      volume_field_names: tuple[str, ...] = tuple(),
      features_module: pytree_transforms.Transform,
      mapping_factory: Callable[..., pytree_mappings.ChannelMapping],
      tendency_transform: pytree_transforms.Transform = pytree_transforms.Identity(),
      input_state_shapes: typing.Pytree | None = None,
      rngs: nnx.Rngs,
  ):
    output_shapes = {}
    (layers,) = coords.vertical.shape
    for field in volume_field_names:
      output_shapes[field] = ShapeFloatStruct(
          (layers,) + coords.horizontal.shape
      )
    for field in surface_field_names:
      output_shapes[field] = ShapeFloatStruct((1,) + coords.horizontal.shape)
    if input_state_shapes is None:
      input_state_shapes = pytree_mappings.MINIMAL_STATE_STRUCT  # default.
    self.explicit_tendency_mapping = mapping_factory(
        input_shapes=features_module.output_shapes(input_state_shapes),
        output_shapes=output_shapes,
        rngs=rngs,
    )
    self.features_module = features_module
    self.tendency_transform = tendency_transform

  def explicit_terms(self, inputs):
    inputs_dict, from_dict_fn = pytree_utils.as_dict(inputs)
    features = self.features_module(inputs_dict)
    tendencies = self.explicit_tendency_mapping(features)
    tendencies = self.tendency_transform(tendencies)
    tendencies = pytree_utils.replace_with_matching_or_default(
        inputs_dict, tendencies, None
    )
    return from_dict_fn(tendencies)


@dataclasses.dataclass
class ComposedODE(ImplicitExplicitODE):
  """Composed equation with exactly one ImplicitExplicitODE instance."""

  equations: Sequence[ExplicitODE | ImplicitExplicitODE]

  def __post_init__(self):
    imex_equations = [
        x for x in self.equations if isinstance(x, ImplicitExplicitODE)
    ]
    if len(imex_equations) != 1:
      raise ValueError(
          'ComposedODE only supports exactly 1 ImplicitExplicitODE, '
          f'got {imex_equations=}'
      )
    implicit_explicit_eq, = imex_equations
    self.implicit_explicit_equation = implicit_explicit_eq

  def explicit_terms(self, x: typing.Pytree) -> typing.Pytree:
    explicit_tendencies = [fn.explicit_terms(x) for fn in self.equations]
    return jax.tree.map(
        lambda *args: sum(x for x in args if x is not None),
        *explicit_tendencies,
        is_leaf=lambda x: x is None,
    )

  def implicit_terms(self, x: typing.Pytree) -> typing.Pytree:
    return self.implicit_explicit_equation.implicit_terms(x)

  def implicit_inverse(
      self, x: typing.Pytree, step_size: float
  ) -> typing.Pytree:
    return self.implicit_explicit_equation.implicit_inverse(x, step_size)


@dataclasses.dataclass
class ComposedExplicitODE(ExplicitODE):
  """Composed explicit equation."""

  equations: Sequence[ExplicitODE]

  def explicit_terms(self, x: typing.Pytree) -> typing.Pytree:
    explicit_tendencies = [fn.explicit_terms(x) for fn in self.equations]
    return jax.tree.map(
        lambda *args: sum([x for x in args if x is not None]),
        *explicit_tendencies,
        is_leaf=lambda x: x is None,
    )



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for converting between different state representations."""

import functools

from dinosaur import primitive_equations as dinosaur_primitive_equations
from dinosaur import spherical_harmonic
from dinosaur import vertical_interpolation
from dinosaur import xarray_utils as dinosaur_xarray_utils
import jax.numpy as jnp

from neuralgcm.experimental import coordinates
from neuralgcm.experimental import equations
from neuralgcm.experimental import orographies
from neuralgcm.experimental import typing
from neuralgcm.experimental import units


def uvtz_to_primitive_equations(
    source_data: dict[str, typing.Array],
    source_coords: coordinates.DinosaurCoordinates,
    primitive_equations: equations.PrimitiveEquations,
    orography,
    sim_units: units.SimUnits,
):
  """Converts velocity/temperature/geopotential to primitive equations state."""
  surface_pressure = vertical_interpolation.get_surface_pressure(
      source_coords.vertical,
      source_data['geopotential'],
      orography.nodal_orography,
      sim_units.g,
  )
  interpolate_fn = vertical_interpolation.vectorize_vertical_interpolation(
      vertical_interpolation.vertical_interpolation
  )
  vertical = primitive_equations.coords.vertical
  assert isinstance(vertical, coordinates.SigmaLevels)
  # TODO(dkochkov): Consider having explicit args that used differently, eg z.
  data_on_sigma = vertical_interpolation.interp_pressure_to_sigma(
      {k: v for k, v in source_data.items() if k != 'geopotential'},
      pressure_coords=source_coords.vertical,
      sigma_coords=vertical.sigma_levels,
      surface_pressure=surface_pressure,
      interpolate_fn=interpolate_fn,
  )
  data_on_sigma['temperature_variation'] = (
      data_on_sigma.pop('temperature') - primitive_equations.T_ref
  )
  data_on_sigma['log_surface_pressure'] = jnp.log(surface_pressure)
  return data_on_sigma


def primitive_equations_to_uvtz(
    source_state: dinosaur_primitive_equations.StateWithTime,
    input_coords: coordinates.DinosaurCoordinates,
    primitive_equations: equations.PrimitiveEquations,
    orography: orographies.Orography,
    target_coords: coordinates.DinosaurCoordinates,
    sim_units: units.SimUnits,
):
  """Converts primitive equations state to pressure level representation.

  This function transforms an atmospheric state described in terms of
  temperature variation, divergence, vorticity, surface pressure and tracers
  on sigma levels to wind components, temperature, geopotential and tracers
  on fixed pressure-level coordinates.

  Args:
    source_state: State in primitive equations representation.
    input_coords: Coordinates of the source state.
    primitive_equations: Primitive equations module.
    orography: Orography module.
    target_coords: Pressure-level dinosaur coordinates to convert to.
    sim_units: Simulation units object.

  Returns:
    Dictionary of state components interpolated to target_coords.
  """
  to_nodal_fn = input_coords.horizontal.ylm_grid.to_nodal
  velocity_fn = functools.partial(
      spherical_harmonic.vor_div_to_uv_nodal,
      input_coords.horizontal.ylm_grid,
  )
  u, v = velocity_fn(  # returned in nodal space.
      vorticity=source_state.vorticity, divergence=source_state.divergence
  )
  t = dinosaur_xarray_utils.temperature_variation_to_absolute(
      to_nodal_fn(source_state.temperature_variation),
      ref_temperature=primitive_equations.T_ref.squeeze())
  tracers = to_nodal_fn(source_state.tracers)
  vertical = input_coords.vertical
  assert isinstance(vertical, coordinates.SigmaLevels)
  z = dinosaur_primitive_equations.get_geopotential_with_moisture(
      temperature=t,
      specific_humidity=tracers['specific_humidity'],
      nodal_orography=orography.nodal_orography,
      coordinates=vertical.sigma_levels,
      gravity_acceleration=sim_units.gravity_acceleration,
      ideal_gas_constant=sim_units.ideal_gas_constant,
      water_vapor_gas_constant=sim_units.water_vapor_gas_constant,
  )
  surface_pressure = jnp.exp(to_nodal_fn(source_state.log_surface_pressure))
  # u, v, t, z, tracers, surface_pressure = (
  #     self.coords.dycore_to_physics_sharding(
  #         (u, v, t, z, tracers, surface_pressure)
  #     )
  # )
  interpolate_with_linear_extrap_fn = (
      vertical_interpolation.vectorize_vertical_interpolation(
          vertical_interpolation.linear_interp_with_linear_extrap
      )
  )
  vertical = primitive_equations.coords.vertical
  assert isinstance(vertical, coordinates.SigmaLevels)
  regrid_with_linear_fn = functools.partial(
      vertical_interpolation.interp_sigma_to_pressure,
      pressure_coords=target_coords.vertical,
      sigma_coords=vertical.sigma_levels,
      surface_pressure=surface_pressure,
      interpolate_fn=interpolate_with_linear_extrap_fn,
  )
  interpolate_with_constant_extrap_fn = (
      vertical_interpolation.vectorize_vertical_interpolation(
          vertical_interpolation.vertical_interpolation
      )
  )
  regrid_with_constant_fn = functools.partial(
      vertical_interpolation.interp_sigma_to_pressure,
      pressure_coords=target_coords.vertical,
      sigma_coords=vertical.sigma_levels,
      surface_pressure=surface_pressure,
      interpolate_fn=interpolate_with_constant_extrap_fn,
  )
  # closest regridding options to those used in ERA5.
  # use constant extrapolation for `u, v, tracers`.
  # use linear extrapolation for `z, t`.
  # google reference: http://shortn/_X09ZAU1jsx.
  outputs = dict(
      u_component_of_wind=regrid_with_constant_fn(u),
      v_component_of_wind=regrid_with_constant_fn(v),
      temperature=regrid_with_linear_fn(t),
      geopotential=regrid_with_linear_fn(z),
      sim_time=source_state.sim_time,
  )
  for k, v in regrid_with_constant_fn(tracers).items():
    outputs[k] = v
  return outputs



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests that random processes generate values with expected stats."""

from absl.testing import absltest
from absl.testing import parameterized
import chex
from flax import nnx
import jax
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import random_processes
from neuralgcm.experimental import typing
from neuralgcm.experimental import units
import numpy as np
import tensorflow_probability.substrates.jax as tfp


@absltest.skipThisClass('Base class')
class BaseSphericalHarmonicRandomProcessTest(parameterized.TestCase):
  """Base class for testing variants of random fields on spherical harmonics."""

  def setUp(self):
    super().setUp()
    self.sim_units = units.DEFAULT_UNITS
    self.dt = self.sim_units.nondimensionalize(typing.Quantity('1 hour'))

  def check_correlation_length(
      self,
      samples,
      expected_correlation_length,
      grid,
  ):
    """Checks the correlation length of random field."""
    unused_n_samples, n_lngs, n_lats = samples.shape
    expected_corr_frac = expected_correlation_length / (2 * np.pi * grid.radius)
    acorr_lat = tfp.stats.auto_correlation(
        # Mean autocorrelation in the lat direction at the longitude=0 line.
        samples[:, 0, :],
        axis=-1,
    ).mean(axis=0)
    # There are 2 * n_lats points in the circumference.
    fractional_corr_len_lat = np.argmax(acorr_lat < 0) / (2 * n_lats)
    self.assertBetween(
        fractional_corr_len_lat,
        expected_corr_frac * 0.5,
        expected_corr_frac * 3,
    )
    acorr_lng = tfp.stats.auto_correlation(
        # Mean autocorrelation in the lng direction at the latitude=0 line.
        samples[:, :, n_lats // 2],
        axis=-1,
    ).mean(axis=0)
    fractional_corr_len_lng = np.argmax(acorr_lng < 0) / n_lngs
    self.assertBetween(
        fractional_corr_len_lng,
        expected_corr_frac * 0.2,
        expected_corr_frac * 3,
    )

  def check_mean(
      self,
      samples,
      grid,
      expected_mean,
      variance,
      correlation_length,
      mean_tol_in_standard_errs,
  ):
    """Checks the mean (at every point & average) of samples."""
    n_samples, unused_n_lngs, unused_n_lats = samples.shape

    # Pointwise mean should be with tol with high probability everywhere.
    # Since we're testing many (lat/lon) points, we allow for some deviations.
    standard_error = np.sqrt(variance) / np.sqrt(n_samples) if variance else 0.0
    np.testing.assert_allclose(
        # 95% of points are within specified tol. There may be outliers.
        np.percentile(np.mean(samples, axis=0), 95),
        expected_mean,
        atol=mean_tol_in_standard_errs * standard_error,
    )
    np.testing.assert_allclose(
        # 100% of points are within a looser tol.
        np.mean(samples, axis=0),
        expected_mean,
        atol=2 * mean_tol_in_standard_errs * standard_error,
    )

    # Check average mean over whole earth (standard_error will be lower so this
    # is a good second check).
    expected_corr_frac = correlation_length / grid.radius
    n_equivalent_integrated_samples = n_samples / expected_corr_frac**2
    if variance:
      standard_error = np.sqrt(variance) / np.sqrt(
          n_equivalent_integrated_samples
      )
    else:
      standard_error = 0.0
    np.testing.assert_allclose(
        np.mean(samples),
        expected_mean,
        atol=mean_tol_in_standard_errs * standard_error,
    )

  def check_variance(
      self,
      samples,
      grid,
      correlation_length,
      expected_variance,
      var_tol_in_standard_errs,
  ):
    expected_variance = expected_variance or 0.0
    expected_integrated_variance = (
        expected_variance * 4 * np.pi * grid.radius**2
    )

    n_samples, unused_n_lngs, unused_n_lats = samples.shape
    # Integrating over the sphere we get additional statistical power since
    # points decorrelate.
    expected_corr_frac = correlation_length / grid.radius
    n_equivalent_integrated_samples = n_samples / expected_corr_frac**2
    standard_error = np.sqrt(
        # The variance of a (normal) variance estimate is 2 σ⁴ / (n - 1).
        2
        * expected_integrated_variance**2
        / n_equivalent_integrated_samples
    )

    np.testing.assert_allclose(
        grid.ylm_grid.integrate(np.var(samples, axis=0)),
        expected_integrated_variance,
        atol=var_tol_in_standard_errs * standard_error,
        rtol=0.0,
    )

  def check_unconditional_and_trajectory_stats(
      self,
      random_field,
      mean,
      variance,
      grid,
      correlation_length,
      correlation_time,
      run_mean_check=True,
      run_variance_check=True,
      run_correlation_length_check=True,
      run_correlation_time_check=True,
      mean_tol_in_standard_errs=4,
      var_tol_in_standard_errs=4,
  ):
    dt = self.dt

    # generating multiple trajectories of random fields.
    n_samples = 500
    unroll_length = 40
    init_rngs = jax.random.split(jax.random.PRNGKey(5), n_samples)
    graph, params = nnx.split(random_field)
    sample_fn = lambda x: graph.apply(params).unconditional_sample(x)[0]
    evaluate_fn = lambda x: graph.apply(params).state_values(grid, x)[0]
    advance_fn = lambda x: graph.apply(params).advance(x)[0]
    batch_sample_fn = jax.vmap(sample_fn)
    batch_evaluate_fn = jax.vmap(evaluate_fn)
    batch_advance_fn = jax.vmap(advance_fn)
    initial_states = batch_sample_fn(init_rngs)
    initial_values = batch_evaluate_fn(initial_states).data

    with self.subTest('unconditional_sample_shape'):
      self.assertEqual(initial_values.shape, (n_samples,) + grid.shape)

    # TODO(dkochkov): Consider adding sizes/size properties to coordinate/field.
    n_lats = grid.fields['latitude'].shape[0]
    n_lngs = grid.fields['longitude'].shape[0]
    self.assertTupleEqual((n_samples, n_lngs, n_lats), initial_values.shape)

    if run_correlation_length_check and variance is not None:
      with self.subTest('unconditional_sample_correlation_len'):
        self.check_correlation_length(initial_values, correlation_length, grid)

    if run_mean_check:
      with self.subTest('unconditional_sample_pointwise_mean'):
        self.check_mean(
            initial_values,
            grid,
            expected_mean=mean,
            variance=variance,
            correlation_length=correlation_length,
            mean_tol_in_standard_errs=mean_tol_in_standard_errs,
        )

    if run_variance_check:
      with self.subTest('unconditional_sample_integrated_var'):
        self.check_variance(
            initial_values,
            grid,
            correlation_length=correlation_length,
            expected_variance=variance,
            var_tol_in_standard_errs=var_tol_in_standard_errs,
        )

    def step_fn(c, _):
      next_c = batch_advance_fn(c)
      next_output = batch_evaluate_fn(next_c).data
      return (next_c, next_output)

    _, field_trajectory = jax.lax.scan(
        step_fn, initial_states, xs=None, length=unroll_length
    )
    field_trajectory = jax.device_get(field_trajectory)

    if run_correlation_time_check and variance is not None:
      with self.subTest('trajectory_correlation_time'):
        # Mean autocorrelation at the lat=lng=0 point.
        acorr = tfp.stats.auto_correlation(
            field_trajectory[:, :, n_lngs // 2, 0], axis=0
        ).mean(axis=1)
        sample_decorr_time = dt * np.argmax(acorr < 0)
        self.assertBetween(
            sample_decorr_time, correlation_time / 2, correlation_time * 2
        )

    final_sample = field_trajectory[-1]

    if run_correlation_length_check and variance is not None:
      with self.subTest('final_sample_correlation_len'):
        self.check_correlation_length(final_sample, correlation_length, grid)

    if run_mean_check:
      with self.subTest('final_sample_pointwise_mean'):
        self.check_mean(
            final_sample,
            grid,
            expected_mean=mean,
            variance=variance,
            correlation_length=correlation_length,
            mean_tol_in_standard_errs=mean_tol_in_standard_errs,
        )

    if run_variance_check:
      with self.subTest('final_sample_integrated_var'):
        self.check_variance(
            final_sample,
            grid,
            correlation_length=correlation_length,
            expected_variance=variance,
            var_tol_in_standard_errs=var_tol_in_standard_errs,
        )

  def check_independent(self, x, y):
    """Checks random field values x and y are independent."""
    self.assertEqual(x.ndim, 2)
    self.assertEqual(y.ndim, 2)
    # corr[i, j] = Correlation(x[:, i], y[:, j])
    corr = tfp.stats.correlation(x, y, sample_axis=0, event_axis=1)
    standard_error = 2 / np.sqrt(x.shape[0])  # product of two iid χ²
    np.testing.assert_array_less(corr, 4 * standard_error)

  def check_nnx_state_structure_is_invariant(self, grf, grid):
    """Checks that random process does not mutate nnx.state(grf) structure."""
    init_nnx_state = nnx.state(grf, nnx.Param)
    random_state = grf.unconditional_sample(jax.random.PRNGKey(0))
    random_state = grf.advance(random_state)
    _ = grf.state_values(grid, random_state)
    nnx_state = nnx.state(grf, nnx.Param)
    chex.assert_trees_all_equal_shapes_and_dtypes(init_nnx_state, nnx_state)


class GaussianRandomFieldTest(BaseSphericalHarmonicRandomProcessTest):
  """Tests GaussianRandomField random process."""

  @parameterized.named_parameters(
      dict(
          testcase_name='T42_reasonable_corrs',
          variance=0.7,
          grid=coordinates.LonLatGrid.T42(),
          correlation_length=0.15,
          correlation_time=3,
      ),
      dict(
          testcase_name='T21_reasonable_corrs',
          variance=1.5,
          grid=coordinates.LonLatGrid.T21(),
          correlation_length=0.15,
          correlation_time=3,
      ),
      dict(
          testcase_name='T42_large_radius',
          variance=1.2,
          grid=coordinates.LonLatGrid.T42(radius=4),
          correlation_length=1.15,
          correlation_time=3,
      ),
      dict(
          testcase_name='T85_long_corrs',
          variance=2.7,
          grid=coordinates.LonLatGrid.T85(),
          correlation_length=0.5,
          correlation_time=3,
      ),
  )
  def test_unconditional_and_trajectory_stats(
      self,
      variance,
      grid,
      correlation_length,
      correlation_time,
  ):
    grf = random_processes.GaussianRandomField(
        grid=grid,
        dt=self.dt,
        sim_units=self.sim_units,
        correlation_time=correlation_time * typing.Quantity('1 hour'),
        correlation_length=correlation_length,
        variance=variance,
        rngs=nnx.Rngs(0),
    )
    self.check_unconditional_and_trajectory_stats(
        grf,
        0.0,  # mean = 0
        variance,
        grid,
        correlation_length,
        correlation_time,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='with_all_nnx_params',
          correlation_time_type=nnx.Param,
          correlation_length_type=nnx.Param,
          variance_type=nnx.Param,
      ),
      dict(
          testcase_name='with_fixed_params',
          correlation_time_type=random_processes.RandomnessParam,
          correlation_length_type=random_processes.RandomnessParam,
          variance_type=random_processes.RandomnessParam,
      ),
      dict(
          testcase_name='with_nnx_and_fixed_params',
          correlation_time_type=nnx.Param,
          correlation_length_type=nnx.Param,
          variance_type=random_processes.RandomnessParam,
      ),
  )
  def test_nnx_state_structure(
      self, correlation_time_type, correlation_length_type, variance_type
  ):
    """Tests that random process does not mutate structure of nnx.state."""
    grid = coordinates.LonLatGrid.T42()
    grf = random_processes.GaussianRandomField(
        grid=grid,
        dt=self.dt,
        sim_units=self.sim_units,
        correlation_time=3 * typing.Quantity('1 hour'),
        correlation_length=0.15,
        variance=1.5,
        correlation_time_type=correlation_time_type,
        correlation_length_type=correlation_length_type,
        variance_type=variance_type,
        rngs=nnx.Rngs(0),
    )
    with self.subTest('nnx_state_structure_invariance'):
      self.check_nnx_state_structure_is_invariant(grf, grid)

    with self.subTest('nnx_param_count'):
      params = nnx.state(grf, nnx.Param)
      actual_count = sum([np.size(x) for x in jax.tree.leaves(params)])
      expected_count = sum(
          x == nnx.Param
          for x in [
              correlation_time_type,
              correlation_length_type,
              variance_type,
          ]
      )
      self.assertEqual(actual_count, expected_count)


class BatchGaussianRandomFieldTest(BaseSphericalHarmonicRandomProcessTest):

  def setUp(self):
    super().setUp()
    self.grid = coordinates.LonLatGrid.T85()

  def _make_grf(
      self,
      variances,
      correlation_lengths,
      correlation_times,
  ):
    return random_processes.BatchGaussianRandomField(
        grid=self.grid,
        dt=self.dt,
        sim_units=self.sim_units,
        correlation_times=correlation_times,
        correlation_lengths=correlation_lengths,
        variances=variances,
        rngs=nnx.Rngs(0),
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='reasonable_corrs',
          variances=(1.0, 2.7),
          correlation_lengths=(0.15, 0.2),
          correlation_times=(1, 2.1),
      ),
  )
  def test_stats(
      self,
      variances,
      correlation_lengths,
      correlation_times,
  ):
    random_field = self._make_grf(
        variances, correlation_lengths, correlation_times
    )
    n_fields = len(variances)
    unroll_length = 10
    n_samples = 1000
    rngs = jax.random.split(jax.random.PRNGKey(802701), n_samples)

    ###
    graph, params = nnx.split(random_field)
    sample_fn = lambda x: graph.apply(params).unconditional_sample(x)[0]
    evaluate_fn = lambda x: graph.apply(params).state_values(self.grid, x)[0]
    advance_fn = lambda x: graph.apply(params).advance(x)[0]
    batch_sample_fn = jax.vmap(sample_fn)
    batch_evaluate_fn = jax.vmap(evaluate_fn)
    batch_advance_fn = jax.vmap(advance_fn)
    initial_states = batch_sample_fn(rngs)
    initial_values = batch_evaluate_fn(initial_states).data

    def step_fn(c, _):
      next_c = batch_advance_fn(c)
      next_output = batch_evaluate_fn(next_c).data
      return (next_c, next_output)

    _, field_trajectory = jax.lax.scan(
        step_fn, initial_states, xs=None, length=unroll_length
    )
    field_trajectory = jax.device_get(field_trajectory)
    ###
    self.assertEqual(
        (unroll_length, n_samples, n_fields) + self.grid.shape,
        field_trajectory.shape,
    )
    final_nodal_value = field_trajectory[-1, ...]

    self.assertEqual(
        (n_samples, n_fields) + self.grid.ylm_grid.modal_shape,
        initial_states.core.shape,
    )

    # Nodal values should have the right statistics.
    for i, (variance, correlation_length) in enumerate(
        zip(variances, correlation_lengths, strict=True)
    ):
      for x in [initial_values, final_nodal_value]:
        self.check_mean(
            x[:, i],
            self.grid,
            expected_mean=0.0,
            variance=variance,
            correlation_length=correlation_length,
            mean_tol_in_standard_errs=5,
        )
        self.check_variance(
            x[:, i],
            self.grid,
            correlation_length=correlation_length,
            expected_variance=variance,
            var_tol_in_standard_errs=5,
        )
        self.check_correlation_length(
            x[:, i],
            expected_correlation_length=correlation_length,
            grid=self.grid,
        )

      # Fields 0 and 1 should be independent.
      # Check a handful of groups of nodal values.
      self.check_independent(
          x[:, 0, 100:105, 60],
          x[:, 1, 100:105, 60],
      )
      self.check_independent(
          x[:, 0, 100:105, 0],
          x[:, 1, 100:105, 0],
      )
      self.check_independent(x[:, 0, 0:5, 0], x[:, 1, 0:5, 0])

    # Initial and final sample should be independent as well, since we unroll
    # for much longer than the correlation time.
    self.check_independent(
        initial_values[:, 0, 50:55, 60],
        final_nodal_value[:, 0, 50:55, 60],
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='with_all_nnx_params',
          correlation_time_type=nnx.Param,
          correlation_length_type=nnx.Param,
          variance_type=nnx.Param,
      ),
      dict(
          testcase_name='with_fixed_params',
          correlation_time_type=random_processes.RandomnessParam,
          correlation_length_type=random_processes.RandomnessParam,
          variance_type=random_processes.RandomnessParam,
      ),
      dict(
          testcase_name='with_nnx_and_fixed_params',
          correlation_time_type=nnx.Param,
          correlation_length_type=nnx.Param,
          variance_type=random_processes.RandomnessParam,
      ),
  )
  def test_nnx_state_structure(
      self, correlation_time_type, correlation_length_type, variance_type
  ):
    """Tests that random process does not mutate structure of nnx.state."""
    grid = coordinates.LonLatGrid.T42()
    grf = random_processes.BatchGaussianRandomField(
        grid=grid,
        dt=self.dt,
        sim_units=self.sim_units,
        correlation_times=(1.0, 2.7),
        correlation_lengths=(0.15, 0.2),
        variances=(1, 2.1),
        correlation_time_type=correlation_time_type,
        correlation_length_type=correlation_length_type,
        variance_type=variance_type,
        rngs=nnx.Rngs(0),
    )
    with self.subTest('nnx_state_structure_invariance'):
      self.check_nnx_state_structure_is_invariant(grf, grid)

    with self.subTest('nnx_param_count'):
      params = nnx.state(grf, nnx.Param)
      actual_count = sum([np.size(x) for x in jax.tree.leaves(params)])
      expected_count = sum(
          grf.n_fields
          for x in [
              correlation_time_type,
              correlation_length_type,
              variance_type,
          ]
          if x == nnx.Param
      )
      self.assertEqual(actual_count, expected_count)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PhysicsSpecifications contain physical parameters of dynamical systems.

To ensure that all model components the expected PhysicsSpecs all modules
(except specializing on a particular equation) must instantiate
PhysicsSpecs objects using `get_physics_specs`, which should be configured
appropriately via `gin`.
"""

from typing import Sequence, Union
from dinosaur import primitive_equations
from dinosaur import scales
from dinosaur import shallow_water
import gin
import numpy as np


# TODO(jamieas): consolidate with `PrimitiveEquationSpecs`. In particular,
# decide whether 'specs' should have units or be nondimensionalized.
QuantityOrStr = Union[str, scales.Quantity]


GET_DEFAULT_SCALE = gin.external_configurable(
    lambda: scales.DEFAULT_SCALE, name='GET_DEFAULT_SCALE')
GET_ATMOSPHERIC_SCALE = gin.external_configurable(
    lambda: scales.ATMOSPHERIC_SCALE, name='GET_ATMOSPHERIC_SCALE')


@gin.configurable
def get_physics_specs(construct_fn=gin.REQUIRED):
  """Returns physical parameters object generated by `construct_fn`."""
  return construct_fn()


@gin.register
def shallow_water_specs_constructor(
    density_vals: Union[Sequence[float], np.ndarray],
    density_units: QuantityOrStr = scales.WATER_DENSITY,
    radius_si: QuantityOrStr = scales.RADIUS,
    angular_velocity_si: QuantityOrStr = scales.ANGULAR_VELOCITY,
    gravity_acceleration_si: QuantityOrStr = scales.GRAVITY_ACCELERATION,
    scale: scales.Scale = scales.DEFAULT_SCALE
) -> shallow_water.ShallowWaterSpecs:
  """Constructs `ShallowWaterSpecs` using gin-configurable parameters.

  Args:
    density_vals: density values for each layer of the shallow water system.
    density_units: units in which `density_vals` are specified.
    radius_si: radius of the domain specified with units attached.
    angular_velocity_si: angular velocity of the domain with units attached.
    gravity_acceleration_si: gravity on the surface with units attached.
    scale: a scale object specifying the scales to use for nondimensionalizing.

  Returns:
    ShallowWaterSpecs object containing physical parameters of the system.
  """
  densities = np.asarray(density_vals) * scales.Quantity(density_units)
  return shallow_water.ShallowWaterSpecs.from_si(
      densities=densities,
      radius_si=scales.Quantity(radius_si),
      angular_velocity_si=scales.Quantity(angular_velocity_si),
      gravity_acceleration_si=scales.Quantity(gravity_acceleration_si),
      scale=scale)


@gin.register
def primitive_eq_specs_constructor(
    radius_si: QuantityOrStr = scales.RADIUS,
    angular_velocity_si: QuantityOrStr = scales.ANGULAR_VELOCITY,
    gravity_acceleration_si: QuantityOrStr = scales.GRAVITY_ACCELERATION,
    ideal_gas_constant_si: QuantityOrStr = scales.IDEAL_GAS_CONSTANT,
    water_vapor_gas_constant_si: QuantityOrStr = scales.IDEAL_GAS_CONSTANT_H20,
    water_vapor_isobaric_heat_capacity_si: QuantityOrStr = (
        scales.WATER_VAPOR_CP),
    kappa_si: QuantityOrStr = scales.KAPPA,
    scale: scales.Scale = scales.DEFAULT_SCALE,
) -> primitive_equations.PrimitiveEquationsSpecs:
  """Constructs `PrimitiveEquationsSpecs` using gin-configurable parameters.

  Args:
    radius_si: radius of the domain with units attached.
    angular_velocity_si: angular velocity of the domain with units attached.
    gravity_acceleration_si: gravity on the surface with units attached.
    ideal_gas_constant_si: the gas constant with units attached.
    water_vapor_gas_constant_si: the gas constant for vapor with units attached.
    water_vapor_isobaric_heat_capacity_si: isobaric heat capacity of vapor with
      units attached.
    kappa_si: `ideal_gas_constant / Cp` where  Cp is the isobaric heat capacity.
    scale: a scale object specifying the scales to use for nondimensionalizing.

  Returns:
    PrimitiveEquationsSpecs object containing physical parameters of the system.
  """
  return primitive_equations.PrimitiveEquationsSpecs.from_si(
      radius_si=scales.Quantity(radius_si),
      angular_velocity_si=scales.Quantity(angular_velocity_si),
      gravity_acceleration_si=scales.Quantity(gravity_acceleration_si),
      ideal_gas_constant_si=scales.Quantity(ideal_gas_constant_si),
      water_vapor_gas_constant_si=scales.Quantity(water_vapor_gas_constant_si),
      water_vapor_isobaric_heat_capacity_si=scales.Quantity(
          water_vapor_isobaric_heat_capacity_si),
      kappa_si=scales.Quantity(kappa_si),
      scale=scale)
